import subprocess
import sys
import os
import inspect
import time
from collections import defaultdict
import json
import csv
import pycanvass.global_variables as gv
import pycanvass.blocks as blocks
import pycanvass.utilities as utilities
from scapy.all import *
import socket
import socketserver
from socketserver import ThreadingTCPServer, StreamRequestHandler
from struct import *
import datetime

global received_datapoints
received_datapoints = 1


class CyberNode:
    """
    A class that represents the cyber capabilities of any node in the network.
    Can be auto-generated in a cyber-physical model, or explicitly generated.
    """

    def __init__(self, name, ip, port, protocol, encryption, CVSS, firewall):
        self.name = name
        self.ip = ip
        self.port = port
        self.protocol = protocol
        self.encryption = encryption
        self.CVSS = CVSS
        self.firewall = firewall


class PairDevices(StreamRequestHandler):
    def handle(self):

        ar2 = []

        for w in range(received_datapoints):
            ar2.append(0.0)
        print(ar2)

        # print ('[i] Ready to stream data from {}'.format(self.client_address))
        conn = self.request
        utilities._data_banner()
        logfile_name = "canvass_cosim_" + str(self.client_address[0]) + "_" + str(self.client_address[1]) + ".csv"
        logfile = open(logfile_name, "w+")

        # metric_file_path = gv.filepaths["metric"]
        # sys.path.insert(0, metric_file_path)
        # temp = os.path.basename(metric_file_path)
        # metric_filename = os.path.splitext(temp)[0]


        while True:
            msg = conn.recv(1024)
            if not msg:
                conn.close()
                print('[i] Disconnected from IP: {}, Port: {}'.format(self.client_address[0], self.client_address[1]))
                sys.exit()

            # print("[i] Querying device: {}".format(self.client_address))

            unpacking_string = '>' + "f" * received_datapoints
            ar = unpack(unpacking_string, msg)
            for c in range(received_datapoints):
                ts = time.time()
                st = datetime.datetime.fromtimestamp(ts).strftime('%Y_%m_%d_%H_%M_%S')
                print("|{:<20}|{:<12}|{:>20}|{:<10}|{:>39}|".format(st, "RECV", self.client_address[0],
                                                             self.client_address[1], ar[c]))
                # if save_data:
                #     log_string = str(time.time()) + ", " + str(self.client_address[0]) + ", " + str(
                #         self.client_address[1]) + ", " + str(ar[c]) + "\n"
                #     logfile.write(log_string)
                # response_from_server = pack('>f', ar[c]*2)
                # print("|{:<20}|{:^12}|{:<20}|{:<10}|{:>39}|".format(time.time(), "SEND" ,self.client_address[0], self.client_address[1], ar[c]*2))
                # conn.send(response_from_server)
            # command_string = 'python ' + gv.filepaths["metric"] + " " + str(ar)
            # os.system(command_string)

            # if ((ar[4] > 5.0) & (ar[4] < 95.0)):
            #     msg2 = pack('>fI', (ar[2] - ar[3]), 1)  # Pin is negative
            # elif (ar[4] < 5.0):
            #     # load shedding
            #     msg2 = pack('>fI', 0.0, 0)
            # else:
            #     # prevent overcharging
            #     msg2 = pack('>fI', 0.0, 1)

            # conn.send(msg2)
            # print("\n")


def get_data(external_device, PORT, datapoints, save_data=False):
    """
    external_device = Name of the client device from which data is to be received.
    PORT = port of the host computer/controller device
    datapoints = Number of data points expected from external device
    """
    server = ThreadingTCPServer(('', PORT), PairDevices)
    received_datapoints = datapoints
    print("[i] Opening Port: {}\n[i] Expecting {} data points\n[?] Waiting for connection from {} >>".format(PORT,
                                                                                                             datapoints,
                                                                                                             external_device))
    server.serve_forever()


# Create a TCP/IP socket
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)


def setup_new_device(cyber_node_name, ip_value, port_value, protocol_name, cvss=None, firewall=False):
    cyber_node = CyberNode(name=cyber_node_name,
                           ip=ip_value,
                           port=port_value,
                           protocol=protocol_name,
                           CVSS=None,
                           firewall=False,
                           encryption=False)
    return cyber_node


def pair_devices(from_device, to_device, protocol, handshake=False, direction="bidirectional"):
    """

    :param from_device:
    :param to_device:
    :param direction:
    :return:
    """

    src_ip = from_device.ip
    src_port = from_device.port
    dst_ip = to_device.ip
    dst_port = to_device.port
    status = False

    try:

        # Configure subprocess to hide the console window
        info = subprocess.STARTUPINFO()
        info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        info.wShowWindow = subprocess.SW_HIDE

        output = subprocess.Popen(['ping', '-n', '1', '-w', '500', str(dst_ip)], stdout=subprocess.PIPE,
                                  startupinfo=info).communicate()[0]

        if "Destination host unreachable" in output.decode('utf-8'):
            print("[i] {} is Offline >> Unreachable".format(to_device.name))
        elif "Request timed out" in output.decode('utf-8'):
            print("[i] {} is Offline >> Timed out".format(to_device.name))
        else:
            print("[i] Connection established between {} and {}".format(to_device.name, from_device.name))
            status = True

        if handshake is True:
            try:
                print("[i] Confirming TCP connection with 3-way handshake")
                hostname = socket.gethostname()
                ip_address_of_host_computer = socket.gethostbyname(hostname)

                sport = random.randint(1024, 65535)
                print("[i] IP Address of host computer = {} : {}".format(ip_address_of_host_computer, sport))
                print("[i] IP Address of destination computer = {} : {}".format(dst_ip, dst_port))
                # SYN
                ip = IP(src=ip_address_of_host_computer, dst=dst_ip)
                SYN = TCP(sport=sport, dport=dst_port, flags='S', seq=1000)
                SYNACK = sr1(ip / SYN)

                # SYN-ACK
                ACK = TCP(sport=sport, dport=dst_port, flags='A', seq=SYNACK.ack + 1, ack=SYNACK.seq + 1)
                send(ip / ACK)
            except:
                print("[i] Check TCPDUMP on Terminal ('netsh' on Windows) or, WireShark")

    except:
        print("[x] Failed to pair {} and {}".format(from_device.name, to_device.name))

    return status


def send_data_packet(to_device, payload="100", protocol="TCP"):
    """
    This function sends data.
    :param name:
    :param payload:
    :param protocol: Default protocol used in TCP/IP
    :return:
    """

    if to_device.ip == "" or to_device.port == "":
        print("[x] Device IP Address or Port is empty, disconnected, or improperly configured.")
        print("[i] Please refer to documentation, or ask Sayon: sayon@ieee.org")
        raise ConnectionError

    if protocol == "TCP":
        packet = IP(dst=to_device.ip / TCP() / Raw(load=payload))
    try:
        sendp(packet)
        print(packet.summary())
    except:
        print("Failed to send data packet.")


# Connect the socket to the port where the server is listening
def connect(device, ip_addr, port, polling_interval=1):
    server_address = (ip_addr, int(port))
    sock.connect(server_address)
    i = 0
    while i < 50:
        sock.sendall("Hello {}. This is Msg {}.\n".format(device, i))
        data = sock.recv(1024)
        if data:
            print("Received from {} --> {}".format(device, data))
            switch_status_dict = json.loads(data)
            for k, v in switch_status_dict.items():
                print("key = {}, value = {}".format(k, v))
        else:
            print("No data received from {}".format(device))
            break
        time.sleep(polling_interval)
        i += 1
    sock.close()


def _print_data_packet(packet):
    print("Received Data Packet")
    print(packet.summary())


def connect_and_control(device, ip_addr, port, polling_interval=1):
    """
    :return:
    """
    server_address = (ip_addr, int(port))
    try:
        sock.connect(server_address)
        sock.sendall("CANVASS CONNECTED: Hello {}!\n".format(device))
    except:
        print("[x] May be {} is not turned on, and set up to send data as a server".format(device))
        return False

    i = 0
    while i < 500000:
        data = sock.recv(1024)
        if data:
            print("Received from {} --> {}".format(device, data))
            sock.sendall("ACK".encode("utf-8"))
            switch_status_dict = json.loads(data)
            for k, v in switch_status_dict.items():
                print("Edge search result = {}".format(_edge_search(k)))
                edit_edge_status(k, flip_status=1)


        else:
            print("No data received from {}".format(device))
            break
        time.sleep(polling_interval)
        i += 1
    sock.close()
    network = blocks.rebuild()
    return network



def _node_search(n):
    n_file = gv.filepaths["nodes"]
    with open(n_file) as f:
        counter = 0
        match_flag = 0
        nodes = csv.reader(f)
        for node in nodes:
            counter += 1
            if node[0].lstrip() == n:
                match_flag = 1
                return counter - 1

        if match_flag == 0:
            called_from = inspect.stack()[1]
            called_module = inspect.getmodule(called_from[0])
            logging.error("[x] Called from - {} : Node {} could not be found.".format(called_module, e))
            return 0


def _edge_search(e):
    e_file = gv.filepaths["edges"]
    with open(e_file) as f:
        counter = 0
        match_flag = 0
        edges = csv.reader(f)
        for edge in edges:
            counter += 1
            if edge[0].lstrip() == e:
                match_flag = 1
                return counter - 1

        if match_flag == 0:
            called_from = inspect.stack()[1]
            called_module = inspect.getmodule(called_from[0])
            logging.error("[x] Called from - {} : Edge {} could not be found.".format(called_module, e))
            return 0


def edit_edge_status(edge_name, set_status=1, availability=1):
    new_rows = []
    e_file = gv.filepaths["edges"]
    edge_search_result = _edge_search(edge_name)
    with open(e_file, 'r+') as f:
        csvr = csv.reader(f)
        csvr = list(csvr)
        for row in csvr:
            new_row = row
            if row[0].lstrip() == edge_name and edge_search_result != 0:
                logging.info("%s : Status %s --> %s, Availability %s --> %s" % (edge_name, row[4], str(set_status), row[13], str(availability)))
                new_row[4] = set_status
                new_row[13] = availability
                # if flip_status == 1:
                #     if row[4].lstrip() == "1":
                #         print("Changing status from ON to OFF.")
                #         new_row[4] = "0"
                #
                #     else:
                #         print("Changing status from OFF to ON.")
                #         new_row[4] = "1"


            new_rows.append(new_row)

    with open(e_file, 'w', newline='') as f:
        # Overwrite the old file with the modified rows
        writer = csv.writer(f)
        writer.writerows(new_rows)



from pycanvass.data_bridge import _node_search
import pycanvass.global_variables as gv
# import pycanvass.resiliency as res
# import pycanvass.blocks as blocks
import re
import csv
import pandas as pd
import logging
# import pycanvass.complexnetworks as cn
import os
from datetime import *


def create_timestamped_node_file(node_to_modify, **kwargs):
    """
    d
    :param timestamp:
    :param node_to_modify:
    :param args:
    :return:
    """
    new_rows = []
    if os.path.isfile('node-file.csv'):
        n_file='node-file.csv'
    else:
        n_file = gv.filepaths["nodes"]

    node_search_result = _node_search(node_to_modify)
    if node_search_result == 0:
        return

    for k, v in kwargs.items():
        if k == "new_load":
            new_load_value = v
        if k == "new_gen":
            new_gen_value = v
    load_key = "new_load"
    gen_key = "new_gen"
    criticality_key = "new_status"
    with open(n_file, 'r+') as f:
        csvr = csv.reader(f)
        csvr = list(csvr)
        for row in csvr:
            new_row = row
            if row[0].lstrip() == node_to_modify.lstrip() or row[9].lstrip() == node_to_modify.lstrip():

                if load_key in kwargs:
                    new_row[5] = kwargs[load_key]

                if gen_key in kwargs:
                    new_row[6] = kwargs[gen_key]

            new_rows.append(new_row)

    new_node_filename = 'node-file.csv'

    with open(new_node_filename, 'w+', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(new_rows)
    full_new_file_path = os.path.join(os.getcwd(), new_node_filename)
    print(full_new_file_path)
    gv.filepaths["nodes"] = full_new_file_path


def create_timestamped_edge_file(timestamp, edge_to_modify, **kwargs):
    """
    d
    :param timestamp:
    :param edge_to_modify:
    :param args:
    :return:
    """
    new_rows = []
    n_file = gv.filepaths["nodes"]

    node_search_result = _node_search(edge_to_modify)
    if node_search_result == 0:
        return

    for k, v in kwargs.items():
        if k == "new_load":
            new_load_value = v
        if k == "new_gen":
            new_gen_value = v
    load_key = "new_load"
    gen_key = "new_gen"
    status_key = "new_status"
    with open(n_file, 'r+') as f:
        csvr = csv.reader(f)
        csvr = list(csvr)
        for row in csvr:
            new_row = row
            if row[0].lstrip() == edge_to_modify.lstrip() or row[9].lstrip() == edge_to_modify.lstrip():

                if load_key in kwargs:
                    new_row[5] = kwargs[load_key]

                if gen_key in kwargs:
                    new_row[6] = kwargs[gen_key]

            new_rows.append(new_row)

    new_node_filename = timestamp + "-temp-node-file.csv"
    logging.info("Creating a temporary load file called {}".format(new_node_filename))


    with open(new_node_filename, 'w+', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(new_rows)

    gv.filepaths["nodes"] = new_node_filename


def search_timestamp_in_df(df, val):
    a = df.index[df['timestamp'].str.contains(val)]
    if a.empty:
        print("[x] No corresponding time series data was found")
        logging.error("Failed at line 115 of eventloop. No corresponding time series data was found")
        raise ValueError
    elif len(a) > 1:
        print("[x] Multiple timestamps found")
        logging.error("Failed at line 119 of eventloop. Multiple corresponding time series data were found")
        raise ValueError
    else:
        # most usual case, only one-time stamp will be found.
        return a.item()


def create_threatfile(threatfile, timestamp):
    df = pd.read_csv(threatfile, skipinitialspace=True)
    fields = list(df)
    threat_anchor_names = []
    df = pd.read_csv(threatfile, skipinitialspace=True, usecols=fields)
    print(df.head())

    row_number = search_timestamp_in_df(df, timestamp)

    with open("threat-file-tmp.csv", "w+") as file:
        file.write("anchor_name,lat,long,strength,water_level,wind_speed,fire_temp,seismic_activity,confidence_factor\n")
        for f in fields:
            if "-wind" in f:
                threat_name = re.split("-wind", f, maxsplit=1)[0]
                wind_risk_index_of_threat = fields.index(f)
                wind_risk_value = df.iloc[row_number, wind_risk_index_of_threat]
                water_risk_value = df.iloc[row_number, wind_risk_index_of_threat+1]
                fire_risk_value = df.iloc[row_number, wind_risk_index_of_threat+2]
                seismic_risk_value = df.iloc[row_number, wind_risk_index_of_threat+3]
                confidence_factor = df.iloc[row_number, wind_risk_index_of_threat+4]
                threat_obj = gv.threat_dict[threat_name]
                write_string = "{},{},{},,{},{},{},{},{}\n".format(threat_obj.anchor, threat_obj.lat, threat_obj.long, water_risk_value, wind_risk_value, fire_risk_value, seismic_risk_value, confidence_factor)
                file.write(write_string)








def timeseries_simulation(start_time="", interval="", vary_load=True, vary_gen=False, vary_criticality=False):

    # For every time step, create a new folder, and a new node file.
    # in that same folder, create a corresponding threat file
    # create a corresponding edge file
    # create a corresponding repair file
    # perform resiliency nodal calculation

    pass



def modify_edge_file(timeseries_dict, vary_availability=True, vary_repairtime=False):
    """

    :param timeseries_dict:
    :param vary_gen:
    :param vary_criticality:
    :return:
    """
    t = datetime.now()
    month_value = "{:02d}".format(t.month)
    simulation_folder_name = "ts_" + str(t.year) + month_value + str(t.day) + str(t.hour) + str(t.minute) + str(t.second)
    logging.info("Time varying edge data will be included")

    df = pd.read_csv(timeseries_dict["edge_availability_curve"], skipinitialspace=True)
    # what columns are there in the file:
    fields = list(df)
    df = pd.read_csv(timeseries_dict["load_profile"], skipinitialspace=True, usecols=fields)
    size = df.shape

    if vary_repairtime is True:
        if "gen" in timeseries_dict:
            df_gen = pd.read_csv(timeseries_dict["generation"], skipinitialspace=True)
        else:
            print("No generation shape timeseries file found.")
            logging.error("No generation shape timeseries file found.")
            continue_sim = input("[?] Do you want to continue the simulation any way without generator data? [Y/N]\n[>] ")
            if continue_sim.lstrip().lower() == "y" or "yes":
                vary_gen = False
            else:
                logging.error("No generation shape timeseries file found. FileNotFoundError will be raised.")
                raise FileNotFoundError

    cwd = os.getcwd()

    newpath = os.path.join(cwd, simulation_folder_name)
    if not os.path.exists(newpath):
        os.mkdir(newpath)
    os.chdir(newpath)

    main_working_dir = os.getcwd()

    for r in range(1, size[1]):
        timestamp = fields[0]
        colname = fields[r]
        for q in range(1, size[0]):
            datetime_obj = datetime.strptime(df[timestamp][q], "%m/%d/%Y %H:%M")
            month_value = "{:02d}".format(datetime_obj.month)
            temp_folder_name = str(datetime_obj.year) + month_value + str(datetime_obj.day) + str(datetime_obj.hour) + str(datetime_obj.minute) + str(datetime_obj.second)
            cwd = os.getcwd()
            newpath = os.path.join(cwd, temp_folder_name)
            if not os.path.exists(newpath):
                os.mkdir(newpath)
            os.chdir(newpath)
            print(str(df[colname][q]))
            if vary_load is True:
                create_timestamped_node_file(colname, new_load=str(df[colname][q]))
            if vary_gen is True:
                create_timestamped_node_file(colname, new_gen=str(df_gen[colname][q]))

            os.chdir(main_working_dir)

def modify_node_file(timeseries_dict, vary_load=True, vary_gen=False, vary_criticality=False):
    """

    :param timeseries_dict:
    :param vary_gen:
    :param vary_criticality:
    :return:
    """
    t = datetime.now()
    month_value = "{:02d}".format(t.month)
    simulation_folder_name = "ts_" + str(t.year) + month_value + str(t.day) + str(t.hour) + str(t.minute) + str(t.second)
    logging.info("timeseries_simulation function is called")

    df = pd.read_csv(timeseries_dict["load_profile"], skipinitialspace=True)
    # what columns are there in the file:
    fields = list(df)
    df = pd.read_csv(timeseries_dict["load_profile"], skipinitialspace=True, usecols=fields)
    size = df.shape

    if vary_gen is True:
        if "gen" in timeseries_dict:
            df_gen = pd.read_csv(timeseries_dict["generation"], skipinitialspace=True)
        else:
            print("No generation shape timeseries file found.")
            logging.error("No generation shape timeseries file found.")
            continue_sim = input("[?] Do you want to continue the simulation any way without generator data? [Y/N]\n[>] ")
            if continue_sim.lstrip().lower() == "y" or "yes":
                vary_gen = False
            else:
                logging.error("No generation shape timeseries file found. FileNotFoundError will be raised.")
                raise FileNotFoundError

    cwd = os.getcwd()

    newpath = os.path.join(cwd, simulation_folder_name)
    if not os.path.exists(newpath):
        os.mkdir(newpath)
    os.chdir(newpath)

    main_working_dir = os.getcwd()

    for r in range(1, size[1]):
        timestamp = fields[0]
        colname = fields[r]
        for q in range(1, size[0]):
            datetime_obj = datetime.strptime(df[timestamp][q], "%m/%d/%Y %H:%M")
            month_value = "{:02d}".format(datetime_obj.month)
            temp_folder_name = str(datetime_obj.year) + month_value + str(datetime_obj.day) + str(datetime_obj.hour) + str(datetime_obj.minute) + str(datetime_obj.second)
            cwd = os.getcwd()
            newpath = os.path.join(cwd, temp_folder_name)
            if not os.path.exists(newpath):
                os.mkdir(newpath)
            os.chdir(newpath)
            print(str(df[colname][q]))
            if vary_load is True:
                create_timestamped_node_file(colname, new_load=str(df[colname][q]))
            if vary_gen is True:
                create_timestamped_node_file(colname, new_gen=str(df_gen[colname][q]))

            # Find the corresponding threat file and copy it here

            os.chdir(main_working_dir)










"""
This library does all the complex network analysis required by the program
"""
# from typing import List, Any, Tuple, Dict, Union
import inspect
import csv
import networkx as nx
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
import json
import random
from pycanvass.all import *
import pycanvass.global_variables as gv
import sys
import pycanvass.utilities as util
import time
import datetime
import os
from pathlib import Path

from networkx import Graph

attempts = 0
settings = ""


def _input_user_pref_file():
    util._banner()
    current_folder_path, current_folder_name = os.path.split(os.getcwd())
    wd = current_folder_path + "\\" + current_folder_name + "\\"
    print("[i] Current working folder: {}".format(wd))
    default_json_file_name = current_folder_path + "\\" + current_folder_name + "\\user_preferences.json"
    default_json_file_name = Path(default_json_file_name)

    if default_json_file_name.exists():
        print("[i] Loaded default user preferences file: {}".format(default_json_file_name))
        # time.sleep(0.5)
        user_preference_path = default_json_file_name
    else:
        print("[x] Default user preferences file not found.")
        user_preference_path = input("\n[i] Please enter the path to your user-preferences file (*.JSON):\n")

    return str(user_preference_path)


while attempts < 4:
    if attempts == 3:
        break
    try:
        user_preference_path = _input_user_pref_file()
        gv.filepaths["user_preferences"] = user_preference_path
        u_settings = open(user_preference_path)
        settings = json.load(u_settings)
        u_settings.close()
        break
    except:
        print("[x] Could not get the user preference file.")
        attempts += 1
        if attempts == 3:
            print("[x] Failed to configure user preferences correctly.")
            sys.exit()
        elif attempts == 2:
            print("[i] One more chance to get the user preference correctly setup.")


# ---------------------------


# # #
# A #
# # #

def add_node_attr(graph):
    """
    Add all the details about the nodes here
    :param graph:
    :return: dictionary containing node features
    """
    all_nodes = graph.nodes()
    n_file = gv.filepaths["nodes"]

    for n in all_nodes:
        node_search_result = _node_search(n)
        try:
            with open(n_file, 'r') as f:
                csvr = csv.reader(f)
                csvr = list(csvr)
                for row in csvr:
                    if row[0].lstrip() == n:
                        if node_search_result is not 0:
                            graph.node[n]['type'] = row[9].lstrip()
                            graph.node[n]['gen'] = row[6].lstrip()
                            graph.node[n]['demand'] = row[5].lstrip()
                            graph.node[n]['lat'] = row[2].lstrip()
                            graph.node[n]['long'] = row[3].lstrip()

        except:
            print("[x] Error in setting up the node attributes. ")
            called_from = inspect.stack()[1]
            called_module = inspect.getmodule(called_from[0])
            logging.error("[x] Called from - {} : Node attributes could not be set.".format(called_module, e))


# # #
# N #
# # #


def _node_search(n):
    n_file = gv.filepaths["nodes"]
    with open(n_file) as f:
        counter = 0
        match_flag = 0
        nodes = csv.reader(f)
        for node in nodes:
            counter += 1
            if node[0].lstrip() == n:
                match_flag = 1
                return counter - 1

        if match_flag == 0:
            called_from = inspect.stack()[1]
            called_module = inspect.getmodule(called_from[0])
            logging.error("[x] Called from - {} : Node {} could not be found.".format(called_module, e))
            return 0


# # #
# B #
# # #


def build_network(node_dict, edge_dict):
    """

    :param node_dict:
    :param edge_dict:
    :return: network_dict: Dictionary containing several network graph objects, and edge details
             network_dict["normal"] : Normal network topology
             network_dict["total"] : Graph containing self-loops, including reconfiguration options
             network_dict["n_open"]: List of all edges which are normally open
             network_dict["n_closed"] : List of all edges which are normally closed
    """

    network_dict = {}

    gv.obj_nodes = []
    gv.obj_edges = []

    for key, value in node_dict.items():
        gv.obj_nodes.append(value)

    for key, value in edge_dict.items():
        gv.obj_edges.append(value)

    graph = nx.Graph()
    total_graph = nx.DiGraph()

    for key, value in node_dict.items():
        gv.obj_nodes.append(value)

    for key, value in edge_dict.items():
        gv.obj_edges.append(value)

    for edge in gv.obj_edges:
        total_graph.add_edge(edge.from_node.lstrip(), edge.to_node.lstrip())
        if edge.status.lstrip() != "0" and edge.availability.lstrip() != "0":
            graph.add_edge(edge.from_node.lstrip(), edge.to_node.lstrip())
            gv.closed_edges_list.append(tuple((edge.from_node.lstrip(), edge.to_node.lstrip())))
        else:
            gv.open_edges_list.append(tuple((edge.from_node.lstrip(), edge.to_node.lstrip())))

    logging.info("Populating nodes and edges with values")

    for n in gv.obj_nodes:
        graph.node[n.name]['name'] = n.name
        graph.node[n.name]['load'] = n.load
        graph.node[n.name]['gen'] = n.gen
        graph.node[n.name]['lat'] = n.lat
        graph.node[n.name]['long'] = n.long
        graph.node[n.name]['kind'] = n.kind
        graph.node[n.name]['critical'] = n.critical

        total_graph.node[n.name]['name'] = n.name
        total_graph.node[n.name]['load'] = n.load
        total_graph.node[n.name]['gen'] = n.gen
        total_graph.node[n.name]['lat'] = n.lat
        total_graph.node[n.name]['long'] = n.long
        total_graph.node[n.name]['kind'] = n.kind
        total_graph.node[n.name]['critical'] = n.critical

        try:
            if n.kind.upper() == "PV" or n.kind.upper() == "SWING":
                gv.all_sources[n.name] = n.gen
            else:
                try:
                    gv.all_loads[n.name] = n.load
                    if int(n.critical) > 5:
                        gv.all_critical_loads[n.name] = n.load
                except ValueError:
                    logging.error("No loads found in the network while making the graph.")
        except ValueError:
            logging.error("No power resource found in the network while making the graph.")

    gv.graph_collection.append(graph)
    gv.graph_collection.append(total_graph)

    for g in gv.graph_collection:
        for e in gv.obj_edges:
            try:
                g[e.from_node.lstrip()][e.to_node.lstrip()]['water_risk'] = e.water_risk
                gv.water_risk_values.append(e.water_risk.lstrip())
            except KeyError:
                logging.info("No edge between %s and %s, so no water risk" % (e.from_node.lstrip(), e.to_node.lstrip()))
                gv.water_risk_values.append(0)

            try:
                g[e.from_node.lstrip()][e.to_node.lstrip()]['wind_risk'] = e.wind_risk
                gv.wind_risk_values.append(e.wind_risk.lstrip())
            except KeyError:
                logging.info(
                    "[i] No edge between %s and %s, so no wind risk " % (e.from_node.lstrip(), e.to_node.lstrip()))
                gv.wind_risk_values.append(0)

            try:
                g[e.from_node.lstrip()][e.to_node.lstrip()]['fire_risk'] = e.fire_risk
                gv.fire_risk_values.append(e.fire_risk.lstrip())
            except KeyError:
                logging.info(
                    "[i] No edge between %s and %s, so no fire risk " % (e.from_node.lstrip(), e.to_node.lstrip()))
                gv.fire_risk_values.append(0)

    network_dict["normal"] = graph
    network_dict["total"] = total_graph
    network_dict["n_open"] = gv.open_edges_list
    network_dict["n_closed"] = gv.closed_edges_list

    return network_dict


# # #
# C #
# # #


def _create_pos_dictionary(graph):
    """
    Internal function, that creates a dictionary that contains position of all nodes in the network
    :param graph:
    :return: dictionary of all nodes with their geo-coordinates,
             dictionary with little shifted position for showing labels
    """
    nodes = graph.nodes()  # list of all nodes <-- keys
    pos_dictionary = {}  # initializing a dictionary to hold the positions
    label_pos_dictionary = {}
    for n in nodes:
        pos_dictionary[n] = tuple((float(graph.node[n]['long']), float(graph.node[n]['lat'])))
        label_pos_dictionary[n] = tuple((float(graph.node[n]['lat']), float(graph.node[n]['long'])))

    return pos_dictionary, label_pos_dictionary


# # #
# F #
# # #

def feeder_path(graph, from_node, to_node):
    """
    :param graph: NetworkX graph object
    :param from_node:
    :param to_node:
    :return: all restoration paths from generator to load nodes
    """
    path_graph = nx.Graph()
    paths = nx.all_simple_paths(graph, from_node, to_node)

    return paths


# # #
# L #
# # #

def lat_long_layout(graph, show=False, save=True, title="Visualization"):
    """
    Help visualize the network graph and some of its resilience properties.
    If your graph does not come with
    :param graph:
    :param show: (Default is False)
    :param save: (Default is True)
    :return:
    """
    focus = "impact_on_edge"
    logging.info(
        "Lat_long_layout function used, focus = {}, trying to display = {}, saving = {}".format(focus, show, save))

    try:
        pos, label_pos = _create_pos_dictionary(graph)
    except ValueError:
        print("[x] Error: Could not create determine where to put the nodes and edges. "
              "... Please check the lat, longs of your nodes. "
              ""
              "... In case your file does not have the lat, longs, generate"
              "... random lat-longs by using the layout_model(<model.glm>, map_random=True) function."
              ""
              "... Please read the documentation on how to do that. ")

        logging.error('Could not create the dictionary for pos, or label_pos')

    color_scheme = settings['visualization']['color_scheme']

    values = []

    for e in graph.edges():
        try:
            values.append(graph[str(e[0])][str(e[1])][focus])
        except:
            values.append(1)

    color_scheme = plt.get_cmap(color_scheme)
    c_norm = colors.Normalize(vmin=0, vmax=7)
    scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=color_scheme)
    color_list = []
    edge_labels = nx.get_edge_attributes(graph, focus)

    for i in range(len(values)):
        color_val = scalar_map.to_rgba(float(values[i]))
        color_list.append(color_val)

    # add a connection to a web visualization

    nx.draw_networkx_edges(graph,
                           pos=pos,
                           edge_color=color_list,
                           # style='dashed',
                           width=2.0)

    nx.draw_networkx_nodes(graph,
                           pos=pos,
                           nodelist=graph.nodes(),
                           node_color='b',
                           node_shape='x')

    nx.draw_networkx_labels(graph,
                            pos=label_pos,
                            edge_labels=edge_labels)

    if show:
        plt.title(title)
        plt.show()

    if save:
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime('%Y_%m_%d_%H_%M_%S')
        filename = "visualization_" + st + ".png"
        plt.savefig(filename)


def loads(critical=False):
    """
    This function lists the loads in the power system network.
    :param critical: Default value is set to False, which means all loads will be displayed in the network.
                     Set this value to True to see only the critical loads
    :return: gv.all_loads or gv.all_critical_loads
    """
    if not critical:
        return gv.all_loads
    else:
        return gv.all_critical_loads


# # #
# M #
# # #


def make_graph(all_edges):
    """

    :param all_edges: List of all instances of Edge objects created by the program
    :return: graph: NetworkX graph object,
             normally_open_edges: number of open edges, that can be closed later if necessary
    """
    graph = nx.Graph()
    total_graph = nx.Graph()

    for edge in all_edges:
        total_graph.add_edge(edge.from_node, edge.to_node)
        if edge.status != "0":
            graph.add_edge(edge.from_node, edge.to_node)
            gv.closed_edges_list.append(tuple((edge.from_node, edge.to_node)))
        else:
            gv.open_edges_list.append(tuple((edge.from_node, edge.to_node)))

    return graph, total_graph, gv.open_edges_list, gv.closed_edges_list


def make_path_graph(nodelist, **kwargs):
    """
    Create a graph based on a found path between a source and a generator
    :param nodelist:
    :return: path_graph: A NetworkX object that is a single path's graph, that can be further used for resiliency
                         analysis

    """

    path_graph = nx.Graph()
    i = 0
    while i < len(nodelist) - 1:
        path_graph.add_edge(nodelist[i], nodelist[i + 1])
        i += 1

    return path_graph


# # #
# P #
# # #


def preprocess(graph, node_dict, edge_dict):
    """

    :param graph:
    :param node_dict:
    :param edge_dict:
    :return:
    """

    gv.obj_nodes = []
    gv.obj_edges = []
    for key, value in node_dict.items():
        gv.obj_nodes.append(value)

    for key, value in edge_dict.items():
        gv.obj_edges.append(value)

    print("[i] Attempting to populate nodes and edges with values")

    for n in gv.obj_nodes:
        graph.node[n.name]['name'] = n.name
        graph.node[n.name]['load'] = n.load
        graph.node[n.name]['gen'] = n.gen
        graph.node[n.name]['lat'] = n.lat
        graph.node[n.name]['long'] = n.long
        graph.node[n.name]['kind'] = n.kind
        graph.node[n.name]['critical'] = n.critical

        try:
            if n.kind.upper() == "PV" or n.kind.upper() == "SWING":
                gv.all_sources[n.name] = n.gen
            else:
                try:
                    gv.all_loads[n.name] = n.load
                    if int(n.critical) > 5:
                        gv.all_critical_loads[n.name] = n.load
                except ValueError:
                    logging.error("No loads found in the network, while making the graph.")
        except ValueError:
            logging.error("No power resource found in the network, while making the graph.")

    for e in gv.obj_edges:
        try:
            graph[e.from_node][e.to_node]['water_risk'] = e.water_risk
            gv.water_risk_values.append(e.water_risk)
        except KeyError:
            print("[i] No edge between %s and %s exist in this topology." % (e.from_node, e.to_node))
            gv.water_risk_values.append(0)

        try:
            graph[e.from_node][e.to_node]['wind_risk'] = e.wind_risk
            gv.water_risk_values.append(e.wind_risk)
        except KeyError:
            print("[i] No edge between %s and %s exist in this topology." % (e.from_node, e.to_node))
            gv.water_risk_values.append(0)

        try:
            graph[e.from_node][e.to_node]['fire_risk'] = e.fire_risk
            gv.water_risk_values.append(e.fire_risk)
        except KeyError:
            print("[i] No edge between %s and %s exist in this topology." % (e.from_node, e.to_node))
            gv.water_risk_values.append(0)


# # #
# R #
# # #


def resiliency_downstream(graph, edgelist, event):
    """
    This function computes the resiliency of all downstream sections,
    mainly looking out for critical loads
    :param graph:
    :param edgelist: All downstream sections of the feeder, from perspective of an edge being analyzed
    :param event: dictionary describing the event magnitude
    :return:
    """

    risk_dict = {}

    for e in edgelist:
        for oe in gv.obj_edges:
            if e[0].lstrip() == oe.from_node.lstrip() and e[1].lstrip() == oe.to_node.lstrip():
                risk_dict[tuple((e[0], e[1]))] = (event["water_risk"] * float(oe.water_risk.lstrip())
                                                  + event["wind_risk"] * float(oe.wind_risk.lstrip())
                                                  + event["fire_risk"] * float(oe.fire_risk.lstrip())) / 30

    print("Risk per edge -->")
    print(risk_dict)

    # loads encountered

    # critical loads encountered

    # generators encountered

    # switches encountered

    pass


def resiliency_upstream(graph, edgelist, event):
    """
    This function computes the resiliency of all upstream sections,
    mainly looking out for generators
    :param edgelist: All upstream sections of the feeder, from persective of a feeder being analyzed
    :return:
    """
    print("Heat your ", edgelist)
    pass


def resiliency(analysis='nodal'):
    """
    This function helps compute the resiliency metric of the network of a
    node, or a network
    :param kwargs:
    :return:
    """

    # 0. what's the event?

    event = gv.event

    # 1. get the graphs

    g1 = gv.graph_collection[0]
    g2 = gv.graph_collection[1]
    nodes = g2.nodes()
    edges = g2.edges()

    # 2. for which node and edge are you doing the analysis?
    wg = nx.DiGraph()  # wg: working graph
    wg.add_edges_from(edges)

    for n in nodes:
        # Finding Downstream Edges
        print("Downstream Edges of %s -->" % n)
        downstream_edges = list(nx.dfs_edges(wg, n))
        resiliency_downstream(g2, downstream_edges, event)

        # Finding Upstream Edges
        print("Upstream Edges of %s -->" % n)
        upstream_edges = list(nx.edge_dfs(wg, n, orientation='reverse'))
        resiliency_upstream(g2, upstream_edges, event)

    # 3. track back from that edge to a source, or multiple sources


# # #
# S #
# # #


def sources():
    """
    This is a querying function that lists the power resources in a network
    :return: list of all power sources within a network
    """
    return gv.all_sources


def simple_show(graph):
    """
    Display the simple network representation of the graph
    :param graph: NetworkX graph object
    :return:
    """
    nx.draw(graph)
    plt.show()


def summary(graph):
    """
    Summarize all the important graph topological properties
    :param graph:
    :return:
    """
    temp_dict = {'degree_centrality': nx.degree_centrality(graph),
                 'betweeness': nx.betweenness_centrality(graph),
                 'resistance_centrality': nx.current_flow_closeness_centrality(graph)}

    return temp_dict

# # #
# V #
# # #



"""
A generic class for edges and nodes, and their status, so that all components can be
manipulated in run-time.
"""

import json
import csv
import pycanvass.global_variables as gv
import os
import pycanvass.complexnetwork as cn
import logging
from shutil import copyfile


def set_simulation_folder(projectname):
    cwd = os.getcwd()
    newpath = os.path.join(cwd, projectname)
    if not os.path.exists(newpath):
        os.mkdir(newpath)


def load_repair():
    """

    :param filename:
    :return:
    """
    repair_file = gv.project["data"]["repair"]

    with open(repair_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)  # Rewind.
        repairs = csv.reader(f)
        if has_header:
            next(repairs)  # Skip header row

        for repair in repairs:
            repairer_name = repair[0]
            repair_obj = Repair(base_name=repair[0],
                                lat=repair[1],
                                long=repair[2],
                                crew=repair[3],
                                pole=repair[4],
                                line=repair[6],
                                transformer=repair[5],
                                handpump=repair[7],
                                switches=repair[8],
                                mobile_genset=repair[9]
                                )

            gv.repair_dict[repairer_name] = repair_obj

    if len(gv.repair_dict) == 0:
        logging.info("No repair crew found")
        print("[i] No repair crew found")
    else:
        logging.info('%d repair crew bases will be used in the simulation' % len(gv.repair_dict))


def load_threats():
    """
    Creates objects of all the threat nodes
    :param filename:
    :return: threat_dict: Dictionary containing all the threat objects.
    """
    threat_file = gv.project["data"]["threats"]

    with open(threat_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)  # Rewind.
        threats = csv.reader(f)
        if has_header:
            next(threats)  # Skip header row

        for threat in threats:
            threat_name = threat[0]
            threat_obj = Threat(anchor=threat[0],
                                lat=threat[1],
                                long=threat[2],
                                strength=threat[3],
                                water_risk=threat[4],
                                fire_risk=threat[6],
                                wind_risk=threat[5],
                                seismic_risk=threat[7],
                                duration=threat[8]
                                )

            gv.threat_dict[threat_name] = threat_obj
    if len(gv.threat_dict) == 0:
        print("[i] No threat anchor points. We will use generic settings and default values.")
        logging.info('No threat anchor points will be used in the simulation.')
    else:
        logging.info('%d threat anchor points will be used in the simulation.' % len(gv.threat_dict))


def rebuild():
    edge_file = gv.project["data"]["edges"]
    gv.filepaths["edges"] = edge_file

    node_file = gv.project["data"]["nodes"]
    gv.filepaths["nodes"] = node_file

    edge_dict = {}
    node_dict = {}

    with open(edge_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)  # Rewind.
        edges = csv.reader(f)
        if has_header:
            next(edges)  # Skip header row
        number = 1
        for edge in edges:
            edge_name = edge[0].lstrip()
            edge_obj = Edge(name=edge[0],
                            kind=edge[1],
                            from_node=edge[2],
                            to_node=edge[3],
                            status=edge[4].lstrip(),
                            r=edge[5], x=edge[6], b=edge[7],
                            fire_risk=edge[8],
                            wind_risk=edge[9],
                            water_risk=edge[10],
                            rating=edge[11],
                            hardening=edge[12])

            edge_dict[edge_name] = edge_obj
            number = number + 1

    # Making the nodes dictionary:

    with open(node_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)
        nodes = csv.reader(f)
        if has_header:
            next(nodes)  # Skip header row

        for node in nodes:
            node_name = node[0].lstrip()
            node_obj = Node(name=node[0],
                            phase=node[1],
                            lat=node[2],
                            long=node[3],
                            voltage=node[4],
                            load=node[5],
                            gen=node[6],
                            kind=node[7],
                            critical=node[8],
                            category=node[9])

            node_dict[node_name] = node_obj

    rebuilt_network_dict = cn.build_network(node_dict, edge_dict)
    return rebuilt_network_dict

def load_project_ts(ts_node_filename, ts_edge_filename, ts_threat_filename,ts_repair_filename):
    """

    :return:
    """
    """
        Creates the JSON object that has all the parameters users can modify to customize the project
        :param filename:
        :return: JSON object
        """
    filename = gv.filepaths["model"]
    project_file = open(filename)
    gv.project = json.load(project_file)
    project_file.close()

    # keep node backup copy
    from shutil import copyfile

    # keep edge backup copy

    # edge_file = gv.project["data"]["edges"]
    edge_file = ts_edge_filename

    try:
        if os.path.isfile(edge_file):
            temp = os.path.basename(edge_file)
            filename = os.path.splitext(temp)[0]
            edge_backup_filename = filename + "-backup.csv"
    except:
        logging.error("Failed to create backup files")

    # node_file = gv.project["data"]["nodes"]
    node_file = gv.filepaths["nodes"]

    metric_file = gv.project["resiliency_metric"]["algorithm"]
    gv.filepaths["metric"] = metric_file

    edge_dict = {}
    node_dict = {}

    set_simulation_folder(gv.project["project_name"])
    # Making the edges dictionary here:

    with open(edge_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)  # Rewind.
        edges = csv.reader(f)
        if has_header:
            next(edges)  # Skip header row
        number = 1
        for edge in edges:
            edge_name = str(edge[0])
            edge_obj = Edge(name=edge[0],
                            kind=edge[1],
                            from_node=edge[2],
                            to_node=edge[3],
                            status=edge[4],
                            r=edge[5], x=edge[6], b=edge[7],
                            fire_risk=edge[8],
                            wind_risk=edge[9],
                            water_risk=edge[10],
                            rating=edge[11],
                            hardening=edge[12],
                            availability=edge[13])

            edge_dict[edge_name] = edge_obj
            number = number + 1

    # Making the nodes dictionary:

    with open(node_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)
        nodes = csv.reader(f)
        if has_header:
            next(nodes)  # Skip header row

        for node in nodes:
            node_name = node[0]
            node_obj = Node(name=node[0],
                            phase=node[1],
                            lat=node[2],
                            long=node[3],
                            voltage=node[4],
                            load=node[5],
                            gen=node[6],
                            kind=node[7],
                            critical=node[8],
                            category=node[9],
                            backup_dg=node[10],
                            wind_cc=node[11],
                            water_cc=node[12],
                            seismic_cc=node[13],
                            fire_cc=node[14],
                            preexisting_damage=node[15],
                            availability=node[16],
                            foliage=node[17],
                            mttr=node[18],
                            op_cost=node[19],
                            repair_cost=node[20]
                            )

            node_dict[node_name] = node_obj

    load_threats()
    load_repair()
    return node_dict, edge_dict


def load_project():
    """
    Creates the JSON object that has all the parameters users can modify to customize the project
    :param filename:
    :return: JSON object
    """
    filename = gv.filepaths["model"]
    project_file = open(filename)
    gv.project = json.load(project_file)
    project_file.close()

    # keep node backup copy
    from shutil import copyfile

    # keep edge backup copy

    edge_file = gv.project["data"]["edges"]
    gv.filepaths["edges"] = edge_file
    try:
        if os.path.isfile(edge_file):
            temp = os.path.basename(edge_file)
            filename = os.path.splitext(temp)[0]
            edge_backup_filename = filename + "-backup.csv"
    except:
        logging.error("Failed to create backup files")

    node_file = gv.project["data"]["nodes"]
    gv.filepaths["nodes"] = node_file

    metric_file = gv.project["resiliency_metric"]["algorithm"]
    gv.filepaths["metric"] = metric_file

    edge_dict = {}
    node_dict = {}

    set_simulation_folder(gv.project["project_name"])
    # Making the edges dictionary here:

    with open(edge_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)  # Rewind.
        edges = csv.reader(f)
        if has_header:
            next(edges)  # Skip header row
        number = 1
        for edge in edges:
            edge_name = str(edge[0])
            edge_obj = Edge(name=edge[0],
                            kind=edge[1],
                            from_node=edge[2],
                            to_node=edge[3],
                            status=edge[4],
                            r=edge[5], x=edge[6], b=edge[7],
                            fire_risk=edge[8],
                            wind_risk=edge[9],
                            water_risk=edge[10],
                            rating=edge[11],
                            hardening=edge[12],
                            availability=edge[13])

            edge_dict[edge_name] = edge_obj
            number = number + 1

    # Making the nodes dictionary:

    with open(node_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)
        nodes = csv.reader(f)
        if has_header:
            next(nodes)  # Skip header row

        for node in nodes:
            node_name = node[0]
            node_obj = Node(name=node[0],
                            phase=node[1],
                            lat=node[2],
                            long=node[3],
                            voltage=node[4],
                            load=node[5],
                            gen=node[6],
                            kind=node[7],
                            critical=node[8],
                            category=node[9],
                            backup_dg=node[10],
                            wind_cc=node[11],
                            water_cc=node[12],
                            seismic_cc=node[13],
                            fire_cc=node[14],
                            preexisting_damage=node[15],
                            availability=node[16],
                            foliage=node[17],
                            mttr=node[18],
                            op_cost=node[19],
                            repair_cost=node[20]
                            )

            node_dict[node_name] = node_obj

    load_threats()
    load_repair()
    return gv.project, node_dict, edge_dict


def make_edges(filename):
    """
    Returns a dictionary of all edges made from available data
    :param filename:
    :return:
    """
    Edges = {}
    with open(filename) as f:
        edges = csv.reader(f)
        for edge in edges:
            edge_name = edge[0].lstrip()
            edge_obj = Edge(name=edge[0],
                            kind=edge[1],
                            from_node=edge[2],
                            to_node=edge[3],
                            status=edge[4],
                            r=edge[5], x=edge[6], b=edge[7],
                            fire_risk=edge[8],
                            wind_risk=edge[9],
                            water_risk=edge[10],
                            rating=edge[11],
                            hardening=edge[12],
                            availability=edge[13])

            Edges[edge_name] = edge_obj

    return Edges


class Repair:
    """
    Threat anchors are the locations where an impact of any event is most likely to be experienced
    """

    def __init__(self, base_name, lat, long, crew, pole, line, transformer, handpump, switches, mobile_genset):
        self.base_name = base_name
        self.lat = lat
        self.long = long
        self.crew = crew
        self.pole = pole
        self.line = line
        self.transformer = transformer
        self.handpump = handpump
        self.switches = switches
        self.mobile_genset = mobile_genset


class Threat:
    """
    Threat anchors are the locations where an impact of any event is most likely to be experienced
    """

    def __init__(self, anchor, lat, long, strength, water_risk, fire_risk, wind_risk, seismic_risk, duration):
        self.anchor = anchor
        self.lat = lat
        self.long = long
        self.strength = strength
        self.water_risk = water_risk
        self.fire_risk = fire_risk
        self.wind_risk = wind_risk
        self.seismic_risk = seismic_risk
        self.duration = duration


class Edge:
    """
    Edges include transformers, overhead Lines, underground Lines, switches, fuses, reclosers, etc.
    """
    numberOfEdges = 0
    allEdges = []

    def __init__(self, name, kind, from_node, to_node, status, r, x, b, wind_risk, water_risk, fire_risk, rating,
                 hardening, availability):
        """
        constructor method
        """
        self.name = name
        self.kind = kind
        self.from_node = from_node
        self.to_node = to_node
        self.status = status  # status: normal status. If Normally Open, status = 1; If Normally Closed, status = 0
        self.r = r  # resistance:
        self.x = x  # impedance:
        self.b = b
        self.wind_risk = wind_risk
        self.water_risk = water_risk
        self.fire_risk = fire_risk
        self.rating = rating
        self.hardening = hardening
        self.availability = availability

        Edge.numberOfEdges += 1
        Edge.allEdges.append(self)

    def count_edges(self):
        """
        Count the number of edges that has been created in the program
        :return:
        """
        print(self.numberOfEdges)


class Node:
    """
    Nodes include all generators, poles, loads, meters, etc.
    """

    numberOfNodes = 0  # Number of all nodes in the network
    allNodes = []  # Collection of all nodes in the network

    def __init__(self, name, phase, lat, long, voltage, load, gen, kind, critical, category, backup_dg, wind_cc,
                 water_cc, seismic_cc, fire_cc, preexisting_damage, availability, foliage, mttr, op_cost, repair_cost):
        self.name = name
        self.phase = phase
        self.lat = lat
        self.long = long
        self.voltage = voltage
        self.load = load
        self.gen = gen
        self.kind = kind
        self.critical = critical
        self.category = category
        self.backup_dg = backup_dg
        self.wind_cc = wind_cc
        self.water_cc = water_cc
        self.seismic_cc = seismic_cc
        self.fire_cc = fire_cc
        self.preexisting_damage = preexisting_damage
        self.availability = availability
        self.foliage = foliage
        self.mttr = mttr
        self.op_cost = op_cost
        self.repair_cost = repair_cost

        Node.numberOfNodes += 1
        Node.allNodes.append(self)

    def count_nodes(self):
        """
        Count the number of nodes that has been created in the program
        :return:
        """
        print(self.numberOfNodes)



"""
Try to discover what happened and why it happened.
"""

from graphviz import Digraph

def create_event_tree():
    print("Create a Fault Tree for Event Analysis")
    name_of_tree = input("[?] Name of analysis")
    dot = Digraph(comment=name_of_tree)
    



import subprocess
import sys
import time
import pandas as pd
"""
All the quick utility functions, frequently called in other libraries
"""

_version = "0.0.2.16"

def reset_data():

    print("[x] Reset data needs to be done manually now.")

def clean_data(string_data):
    clean_data = string_data.replace(',', '')
    clean_data = clean_data.lstrip()
    clean_data = clean_data.replace('(', '').replace(')', '')
    return clean_data

def _merge_and_align_columns(file_a, file_b):
    df_main = pd.read_csv(file_a)
    print(df_main.loc['lat'])
    df_short = pd.read_csv(file_b)
    return 0

def _data_banner():
    """
    Displays just before stream data is shown on the node screen.
    """
    print("|=========================================================================================================|")
    print("| CANVASS Co-Sim Module 0.0.1                                                                             |")
    print("|---------------------------------------------------------------------------------------------------------|")
    print("| Starting Time: {} seconds, since Jan 1, 1970.".format(time.time()))
    print("| Questions? Ask: Sayonsom Chanda, Email: sayon@ieee.org                                                  |")
    print("|====================|============|====================|==========|=======================================|")
    print("|TIME STAMP          | Direction  | DEVICE IP          | PORT     | DATA RECEIVED                         |")
    print("|====================|============|====================|==========|=======================================|")

def _hide_terminal_output():
    """
    Configure subprocess to hide the console window
    """
    info = subprocess.STARTUPINFO()
    info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    info.wShowWindow = subprocess.SW_HIDE


def _update_progress(progress):
    barLength = 10 # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength*progress))
    text = "\rPercent: [{0}] {1}% {2}".format( "#"*block + "-"*(barLength-block), progress*100, status)
    sys.stdout.write(text)
    sys.stdout.flush()

def _banner():

    print("")
    print("  /$$$$$$                                                            ")
    print(" /$$__  $$                                                           ")
    print("| $$  \__/  /$$$$$$  /$$$$$$$  /$$    /$$ /$$$$$$   /$$$$$$$ /$$$$$$$")
    print("| $$       |____  $$| $$__  $$|  $$  /$$/|____  $$ /$$_____//$$_____/")
    print("| $$        /$$$$$$$| $$  \ $$ \  $$/$$/  /$$$$$$$|  $$$$$$|  $$$$$$ ")
    print("| $$    $$ /$$__  $$| $$  | $$  \  $$$/  /$$__  $$ \____  $$\____  $$")
    print("|  $$$$$$/|  $$$$$$$| $$  | $$   \  $/  |  $$$$$$$ /$$$$$$$//$$$$$$$/")
    print(" \______/  \_______/|__/  |__/    \_/    \_______/|_______/|_______/ ")
    print("                                                                     ")
    print("")
    print("Cyber Attack & Network Vulnerability Analytics Software for Smartgrids")
    print("----------------------------------------------------------------------")
    print("Version           :	   {}".format(_version))
    print("Author            :     Sayonsom Chanda")
    print("License           :     GNU v2")
    print("Help & How to     :     sayon@ieee.org")
    print("----------------------------------------------------------------------")          

def _banner_2(): 
    print("\n" * 100)
    print("|----------------------------------------------|")
    print("| pyCanvass: {}                          |".format(_version))
    print("|----------------------------------------------|")
    print("| Resiliency computation tool for Smart Grids  |")
    print("|                                              |")
    print("| Author: Sayonsom Chanda                      |")
    print("| License: GNU Public License, v2: 2016-18     |")
    print("| Support: sayon@ieee.org                      |")
    print("|----------------------------------------------|")



import numpy as np
import pandas as pd
import matplotlib as mpl
from pathlib import Path
import networkx as nx
import json
import pycanvass.global_variables as gv
# from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
from matplotlib.mlab import griddata
import logging
import subprocess
import sys
import re
import os
import pycanvass.utilities as util
import random
from pycanvass.all import *

# SET PLOT VISUALIZATION DEFAULTS
# -------------------------------
SMALL_FONT = 8
MEDIUM_FONT = 12
LARGE_FONT = 20

plt.rc('figure', titlesize=LARGE_FONT)
u_settings = open(gv.filepaths["user_preferences"])
settings = json.load(u_settings)
u_settings.close()
interpolation = settings['visualization']['interpolation']
# -------------------------------

def _grid(x, y, z, resX=100, resY=100):
    xi = np.linspace(min(x), max(x), resX)
    yi = np.linspace(min(y), max(y), resY)
    Z = griddata(x, y, z, xi, yi, interp=interpolation)
    X, Y = np.meshgrid(xi, yi)
    return X, Y, Z


def view_on_a_map(filename, **kwargs):
    """
    Plots a feeder file on a map
    :param filename:
    :param kwargs:
    :return:
    """
    if kwargs is not None:
        for key, value in kwargs.iteritems():
            if key == "location":
                location = value
            elif key == "location_latlong":
                location_latlong = value
            elif key == "radius":
                radius = value
            elif key == "basemap":
                basemap = value


def _next_geo_coordinate(geo_coord, distance):
    """
    Calculates the next geo-coordinate.
    If we know one geo-coordinate and how far the coordinate is supposed to be, we can make a random guess about the next co-ordinate point.
    :param geo_coord: tuple containing [lat, long]
    :param distance: in feet.
    :return: next_geo_coord tuple containing [lat, long]

    Note:
    - 1 deg latitude = 111.32 km, or 69 miles, or 364320 feet
    - 1 ft = 0.0003048 km, or 2.7448e-6 latitude shift

    """
    theta = random.randrange(0, 359)
    # direction = ""
    # if theta >= 0 and theta < 30:
    #     direction = "E"
    # elif theta >= 30 and theta < 60:
    #     direction = "NE"
    # elif theta >= 60 and theta < 120:
    #     direction = "N"
    # elif theta >= 120 and theta < 150:
    #     direction = "NW"
    # elif theta >= 150 and theta < 210:
    #     direction = "W"
    # elif theta >= 210 and theta < 240:
    #     direction = "SW"
    # elif theta >= 240 and theta < 300:
    #     direction = "S"
    # elif theta >= 300 and theta < 330:
    #     direction = "SE"
    # elif theta >= 330 and theta < 360:
    #     direction = "E"

    # print("theta {} : direction {}".format(theta, direction))
    if distance != "None":
        next_lat = geo_coord[0] + float(distance) * np.cos(theta * np.pi / 180) * 2.7448e-6
        next_long = geo_coord[1] + float(distance) * np.sin(theta * np.pi / 180) * 2.7448e-6
        logging.info("_next_geo_coordinate calculated lat: {} long: {}".format(next_lat, next_long))
        return [next_lat, next_long]

def threat_graph():
    fields = ['anchor_name', 'lat', 'long', 'strength']
    df = pd.read_csv("threat-anchor-file.csv", skipinitialspace=True, usecols=fields)
    nodes = df.anchor_name
    threat_graph = nx.path_graph(nodes)
    threat_graph_pos = {}
    i = 0
    for n in nodes:
        threat_graph.add_node(n, pos=(df.long[i], df.lat[i]))
        threat_graph_pos[n] = tuple((df.long[i], df.lat[i]))
        i += 1
    return threat_graph, threat_graph_pos

def visualize(mode="contour_plot", graph=None, criteria="node_risk", title=""):
    """

    :param mode: (Default: mode="contour_plot") Other options are "bar_plot", "line_chart")
    :param model_name:
    :param criteria: (Default: risk) Other option is resiliency.
    :return:
    """

    project_config_file = open(gv.filepaths["model"])
    project_settings = json.load(project_config_file)
    project_config_file.close()

    if criteria == "node_risk":
        expected_file_name = project_settings["project_name"] + "-nodal_calculation.csv"
        expected_file_name = Path(expected_file_name)

    if expected_file_name.exists():
        # RISK PLOTS
        util._hide_terminal_output()
        fields = ['lat', 'long', 'risk']
        logging.info("[i] Plotting contour plot of the {} of {}".format(criteria, project_settings["project_name"]))
        df = pd.read_csv(expected_file_name, skipinitialspace=True, usecols=fields)
        x = df.long
        y = df.lat
        z = df.risk
        X, Y, Z = _grid(x, y, z)
        # fig = plt.figure()
        # ax = fig.gca()
        # ax.set_xticks(np.arange(min(x), max(x)))
        # ax.set_yticks(np.arange(min(y), max(y)))
        plt.grid()
        plt.contourf(X, Y, Z)
        if title == "":

            plt.title("{} of {} with respect to {} of intensity {}".format(criteria, project_settings["project_name"],
                                                                           project_settings["event"]["type"],
                                                                           project_settings["event"][
                                                                               "known_intensity"]))
        else:
            plt.title(title)


        if graph is not None:
            pos, label_pos = cn._create_pos_dictionary(graph)
            nx.draw_networkx_edges(graph,
                                   pos=pos,
                                   # style='dashed',
                                   width=2.0)

            nx.draw_networkx_nodes(graph,
                                   pos=pos,
                                   nodelist=graph.nodes(),
                                   node_color='b',
                                   node_shape='>')

            tg, tg_pos = threat_graph()
            nx.draw_networkx_nodes(tg,pos=tg_pos, alpha=0.75, node_size=200, node_shape="x")
            nx.draw_networkx_edges(tg,pos=tg_pos, edge_color='r', alpha=0.75, style='dashed',width=2.5)
        plt.colorbar()
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")

        plt.show()


    else:
        print("[x] Could not find the default risk and resiliency file name for {}".format(expected_file_name))
        print("[i] Please read the documentation on naming conventions")
        logging.error("Couldn't find the default risk and resiliency file name for {}.".format(expected_file_name))


def layout_model(file_or_folder_name, map_random=False):
    """
    Create a tree like layout for a feeder model. The valid files that you can supply are the edge csv file (*.csv),
     or the feeder model Gridlab-d file (*.glm).
    This is a memory intensive operation, with O(n^3) running time. <-- todo
    Please be patient if you use this operation.

    Caveat: This function does not check if the GridLAB-D model converges or not, so the user is responsible for
    checking model accuracy.

    :param map_random:
    :return:
    :param file_or_folder_name: [1] a GridLAB model file (i.e. file with *.glm extension)
    or, [2] a folder containing multiple GridLAB-D model.

    :return: SVG file (or files) containing the layout of the GridLAB-D Model
    """
    try:
        util._hide_terminal_output()

        if not os.path.exists('C:\\Program Files (x86)\\Graphviz2.38\\bin'):
            print("[x] Dependency Error: Graphviz is not installed.")
            logging.error("Dependency Error: Graphviz is not installed.")
            print("[i] Layout of the GridLAB-D Models will not proceed.")
            print("[i] Please refer to documentation on how to get Graphviz, and set path correctly")
            return
    except Exception:

        print("[i] Please refer to documentation")
        logging.error("Error in layout_model -- Tried to search if GraphViz is installed, and failed.")
        return

    if os.path.isfile(file_or_folder_name):
        temp = os.path.basename(file_or_folder_name)
        filename = os.path.splitext(temp)[0]
        file_extension = os.path.splitext(temp)[1]

        if 'edge' in filename and file_extension == ".csv":
            logging.info("Will create lay out model using the edge csv file only.")
            filename = filename + "-simple"
            outfile = open(filename, 'w')
            outfile.write("digraph {\n")
            e_file = gv.filepaths["edges"]

            with open(e_file, 'r+') as f:
                csvr = csv.reader(f)
                csvr = list(csvr)[1:]
                for row in csvr:
                    # if row[2].lstrip() is not "from_node":
                    if row[1].lstrip() == "switch":
                        if row[4].lstrip() == "0":
                            outfile.write('\t{} -> {} [color=red, style=dotted]\n'.format(row[2], row[3]))
                        else:
                            outfile.write('\t{} -> {} [color=red]\n'.format(row[2], row[3]))
                    else:
                        outfile.write('\t{} -> {}\n'.format(row[2], row[3]))
                outfile.write('}\n')
                outfile.close()
                os.system("dot -Tsvg -O " + filename)

        elif file_extension == ".glm":
            logging.info("Will create lay out model using the GridLAB-D Feeder model.")

            infile = open(file_or_folder_name, 'r')
            outfile = open(filename, 'w')

            e_file = gv.filepaths["edges"]

            lines = infile.readlines()

            # .dot files begin with the 'graph' or 'digraph' keyword.
            outfile.write("digraph {\n")
            outfile.write("node [shape=box]\n")
            # These are state variables.  There is no error checking, so we rely on
            # well formatted *.GLM files.
            s = 0
            state = 'start'
            edge_color = 'black'
            edge_style = 'solid'
            lengthVal = 0.0
            lineLengthIncrement = 0

            written_nodes = []
            last_node_traversed = []
            if map_random is True:
                node_latlong_file_name = filename + "-node-latlong-file.csv"
                node_latlong_outfile = open(node_latlong_file_name, 'w+')
                node_latlong_outfile.write("node,lat,long,risk\n")

            # Loop through each line in the input file...
            while s < len(lines):
                # Discard Comments
                if re.match("//", lines[s]) is None:
                    if re.search("from", lines[s]) is not None:
                        ts = lines[s].split()
                        # Graphviz format can't handle '-' characters, so they are converted to '_'
                        ns = ts[1].rstrip(';').replace('-', '_').replace(':', '_')
                        outfile.write(ns)

                        if len(written_nodes) == 0:
                            print(
                                "\n\n[i] Provide substation latitude and longitude (N,E --> +ve, W,S --> -ve)\n... Rest of the network geo-coordinates are randomly populated.")
                            starting_lat = input("[?] Latitude   : ")
                            starting_long = input("[?] Longitude  : ")
                            written_nodes.append(ns)
                            node_latlong_outfile.write(ns + "," + str(starting_lat) + "," + str(starting_long) + "\n")
                            last_node_traversed = [ns, float(starting_lat), float(starting_long)]
                            logging.info(
                                "Starting to calculate distance from Lat: {} Long: {}".format(float(starting_lat),
                                                                                              float(starting_long)))

                        state = 'after_from'
                    elif state == 'after_from' and re.search("to ", lines[s]) != None:
                        ts = lines[s].split()
                        ns = ts[1].rstrip(';').replace('-', '_').replace(':', '_')
                        if edge_color == 'red':
                            outfile.write(
                                ' -> ' + ns + '[style=' + edge_style + ' color=' + edge_color + ' label="' + lengthVal + '"]\n')
                            outfile.write("node [shape=box]\n")
                            logging.info(
                                "if_loop: Calculating distance from {} to {}, which is Lat: {} Long: {}".
                                    format(last_node_traversed[0], ns, last_node_traversed[1], last_node_traversed[2]))
                            x = _next_geo_coordinate([last_node_traversed[1], last_node_traversed[2]],
                                                     float(lengthVal2))
                            if ns not in written_nodes:
                                node_latlong_outfile.write(ns + "," + str(x[0]) + "," + str(x[1]) + "\n")
                                written_nodes.append(ns)
                            last_node_traversed[0] = ns
                            last_node_traversed[1] = float(x[0])
                            last_node_traversed[2] = float(x[1])

                        else:
                            outfile.write(
                                ' -> ' + ns + '[style=' + edge_style + ' color=' + edge_color + ' label="' + lengthVal2 + '"]\n')
                            logging.info(
                                "else_loop: Calculating distance from {} to {}, which is Lat: {} Long: {}".
                                    format(last_node_traversed[0], ns, last_node_traversed[1], last_node_traversed[2]))
                            x = _next_geo_coordinate([last_node_traversed[1], last_node_traversed[2]],
                                                     float(lengthVal2))
                            if ns not in written_nodes:
                                node_latlong_outfile.write(ns + "," + str(x[0]) + "," + str(x[1]) + "\n")
                                written_nodes.append(ns)

                            last_node_traversed[0] = ns
                            last_node_traversed[1] = float(x[0])
                            last_node_traversed[2] = float(x[1])

                        lengthVal = 0.0
                        # After an edge is added to the graph, reset the states back to default
                        state = 'start'
                        edge_color = 'black'
                        edge_style = 'solid'
                    elif (re.search("object underground_line", lines[s]) is not None) or (
                            re.search("object overhead_line", lines[s]) is not None):
                        while '}' not in lines[s + lineLengthIncrement]:
                            if re.search("length ", lines[s + lineLengthIncrement]) is not None:
                                les = lines[s + lineLengthIncrement].split()
                                if len(les) > 2:
                                    tsVal = les[1] + ' ' + les[2].strip(';')
                                    lengthVal = tsVal
                                elif len(les) <= 2:
                                    tsVal = les[1].strip(';')
                                    lengthVal = tsVal
                                break
                            else:
                                lineLengthIncrement = lineLengthIncrement + 1
                            if '}' in lines[s + lineLengthIncrement]:
                                lineLengthIncrement = 0
                                lengthVal = 0.0
                                break
                        if re.search("object underground_line", lines[s]) is not None:
                            lengthVal2 = lengthVal
                            lengthVal = "UG_line\\n" + str(lengthVal)

                        elif re.search("object overhead_line", lines[s]) is not None:
                            lengthVal2 = lengthVal
                            lengthVal = "OH_line\\n" + str(lengthVal)
                    elif re.search("object transformer", lines[s]) is not None:
                        edge_color = 'red'
                        lengthVal2 = "0.0"
                        lengthVal = "transformer\\n" + str(lengthVal)
                        outfile.write("node [shape=oval]\n")
                    elif re.search("object triplex_line", lines[s]) is not None:
                        edge_color = 'green'
                        lengthVal2 = lengthVal
                        lengthVal = "triplex_line\\n" + str(lengthVal)
                    elif re.search("object fuse", lines[s]) != None:
                        edge_color = 'blue'
                        lengthVal2 = lengthVal
                        lengthVal = "Fuse\\n" + str(lengthVal)
                    elif re.search("phases ", lines[s]) != None:
                        ts = lines[s].split()
                        if len(ts[1].rstrip(';')) > 3:
                            edge_style = 'bold'
                        elif len(ts[1].rstrip(';')) > 2:
                            edge_style = 'dashed'

                s += 1

            outfile.write("}\n")
            infile.close()
            outfile.close()
            os.system("dot -Tsvg -O " + filename)

        else:
            print("[x] {}.{} is not a GridLAB-D File or valid Edge csv file. Layout operation is skipped.".format(
                filename, file_extension))
            logging.debug(
                "{}.{} is not a GridLAB-D File or valid Edge csv file. Layout operation is skipped.".format(filename,
                                                                                                            file_extension))


    elif os.path.isdir(file_or_folder_name):
        print("[i] Creating a tree layout for all GLM files in: {}".format(file_or_folder_name))
        logging.info("Creating a tree layout all GLM files in: {}".format(file_or_folder_name))
        for f in os.listdir(file_or_folder_name):
            filename = os.path.join(file_or_folder_name, f)
            print("[i] Processing: {}".format(filename))
            logging.debug("Processing: {}".format(filename))
            layout_model(filename)



    else:
        print(
            "[i] The parameter {} you used inside `layout_model({})` was not found to be either file or a directory.".format(
                file_or_folder_name, file_or_folder_name))
        print("[i] Please re-check the parameter. If error persists, check the Known Issues document.")
        logging.error(
            "The parameter {} you used inside `layout_model({})` was not found to be either file or a directory.".format(
                file_or_folder_name, file_or_folder_name))
        return

    print("[i] Layout of all GridLAB-D files are complete.")
    logging.info("Layout of all GridLAB-D files are complete.")
    print("[i] SVG files can be viewed in any web browser, and edited using Inkscape.")

    return


def realtime_demo():
    """
    """
    plt.axis([0, 10, 0, 1])

    for i in range(10):
        y = np.random.random()
        plt.scatter(i, y)
        plt.pause(0.05)

    plt.show()



import pycanvass.utilities as util

import sys
import time

demand_rt = sys.argv[1].replace(',','')
demand_rt = demand_rt.lstrip()
demand_rt = demand_rt.replace('(','').replace(')','')

gen_rt = sys.argv[2].replace(',','')
demand_rt = gen_rt.lstrip()
demand_rt = gen_rt.replace('(','').replace(')','')

for arg, in sys.argv:
    util.clean_data(arg)




def later_metric(graph, node_file, demand_rt, demand_now, t_event):
    """
    This metric was developed as a part of GMLC 1_3_9 project.
    :param graph:
    :param node_file:
    :param demand_rt:
    :param demand_now:
    :param t_event:
    :return:
    """

    n_file = gv.filepaths["nodes"]
    all_nodes = graph.nodes()
    all_node_metrics = {}
    for n in all_nodes:
        node_search_result = _node_search(n)

        try:
            with open(n_file, 'r') as f:
                csvr = csv.reader(f)
                csvr = list(csvr)
                for row in csvr:
                    if row[0].lstrip() == n:
                        if node_search_result is not 0:
                            demand_pre = graph.node[n]['demand'] = row[5].lstrip()
                            t_now = time.time()
                            metric = (row[8] * demand_now / demand_pre) / (t_now - t_event)
                            all_node_metrics[n] = metric
        except:
            print("Error, bye bye")
            sys.exit()

    return all_node_metrics



"""
This file contains all the frequently used global variables
used by other parts of the program

author: Sayonsom Chanda
version: 0.0.4
package: PyCanvass
last_modified: 05/16/2018

"""

global filepaths
filepaths = {'nodes': "", 'edge': "", 'model': "", "user_preferences": "", 'metric': "", "backup": {'nodes': "", 'edge': ""}}

global data_from_device
data_from_device = {}

global event
event = {"water_risk": 0.069, "wind_risk": 0.1, "fire_risk": 1, "seismic_risk" : 1.0}

global node_risk_dict
node_risk_dict = {}

global edge_risk_dict
edge_risk_dict = {}

global wind_risk_values
wind_risk_values = []

global water_risk_values
water_risk_values = []

global fire_risk_values
fire_risk_values = []

global cyber_risk_values
cyber_risk_values = []

global obj_nodes
obj_nodes = []

global obj_edges
obj_edges = []

global project
project = {}

global open_edges_list
open_edges_list = []

global closed_edges_list
closed_edges_list = []

global threat_dict
threat_dict = {}

global repair_dict
repair_dict = {}

global all_sources
all_sources = {}

global all_loads
all_loads = {}

global all_critical_loads
all_critical_loads = {}

global graph_collection
graph_collection = []

global node_res
node_res = {}

global node_biz
node_biz = {}

global node_law
node_law = {}

global node_shl
node_shl = {}

global node_sup
node_sup = {}

global node_wat
node_wat = {}

global node_com
node_com = {}

global node_trn
node_trn = {}

global node_wlf
node_wlf = {}

global node_utl
node_utl = {}




"""
Computes minimum spanning tree of a weighted graph.

"""

__all__ = ['kruskal_mst',
           'minimum_spanning_edges',
           'minimum_spanning_tree',
           'prim_mst_edges', 'prim_mst']

import networkx as nx
from heapq import heappop, heappush
from itertools import count


def minimum_spanning_edges(G, weight='weight', data=True):
    """Generate edges in a minimum spanning forest of an undirected
    weighted graph.

    A minimum spanning tree is a subgraph of the graph (a tree)
    with the minimum sum of edge weights.  A spanning forest is a
    union of the spanning trees for each connected component of the graph.

    Parameters
    ----------
    G : NetworkX Graph

    weight : string
       Edge data key to use for weight (default 'weight').

    data : bool, optional
       If True yield the edge data along with the edge.

    Returns
    -------
    edges : iterator
       A generator that produces edges in the minimum spanning tree.
       The edges are three-tuples (u,v,w) where w is the weight.


    Notes
    -----
    Uses Kruskal's algorithm.

    """

    from networkx.utils import UnionFind
    if G.is_directed():
        raise nx.NetworkXError(
            "Mimimum spanning tree not defined for directed graphs.")

    subtrees = UnionFind()
    edges = sorted(G.edges(data=True), key=lambda t: t[2].get(weight, 1))
    for u, v, d in edges:
        if subtrees[u] != subtrees[v]:
            if data:
                yield (u, v, d)
            else:
                yield (u, v)
            subtrees.union(u, v)


def minimum_spanning_tree(G, weight='weight'):
    """Return a minimum spanning tree or forest of an undirected
    weighted graph.

    A minimum spanning tree is a subgraph of the graph (a tree) with
    the minimum sum of edge weights.

    If the graph is not connected a spanning forest is constructed.  A
    spanning forest is a union of the spanning trees for each
    connected component of the graph.

    Parameters
    ----------
    G : NetworkX Graph

    weight : string
       Edge data key to use for weight (default 'weight').

    Returns
    -------
    G : NetworkX Graph
       A minimum spanning tree or forest.

    Uses Kruskal's algorithm.

    If the graph edges do not have a weight attribute a default weight of 1
    will be used.
    """
    T = nx.Graph(nx.minimum_spanning_edges(G, weight=weight, data=True))
    # Add isolated nodes
    if len(T) != len(G):
        T.add_nodes_from([n for n, d in G.degree().items() if d == 0])
    # Add node and graph attributes as shallow copy
    for n in T:
        T.node[n] = G.node[n].copy()
    T.graph = G.graph.copy()
    return T

kruskal_mst = minimum_spanning_tree


def prim_mst_edges(G, weight='weight', data=True):
    """Generate edges in a minimum spanning forest of an undirected
    weighted graph.

    A minimum spanning tree is a subgraph of the graph (a tree)
    with the minimum sum of edge weights.  A spanning forest is a
    union of the spanning trees for each connected component of the graph.

    Parameters
    ----------
    G : NetworkX Graph

    weight : string
       Edge data key to use for weight (default 'weight').

    data : bool, optional
       If True yield the edge data along with the edge.

    Returns
    -------
    edges : iterator
       A generator that produces edges in the minimum spanning tree.
       The edges are three-tuples (u,v,w) where w is the weight.


    Notes
    -----
    Uses Prim's algorithm.

    If the graph edges do not have a weight attribute a default weight of 1
    will be used.
    """

    if G.is_directed():
        raise nx.NetworkXError(
            "Mimimum spanning tree not defined for directed graphs.")

    push = heappush
    pop = heappop

    nodes = G.nodes()
    c = count()

    while nodes:
        u = nodes.pop(0)
        frontier = []
        visited = [u]
        for u, v in G.edges(u):
            push(frontier, (G[u][v].get(weight, 1), next(c), u, v))

        while frontier:
            W, _, u, v = pop(frontier)
            if v in visited:
                continue
            visited.append(v)
            nodes.remove(v)
            for v, w in G.edges(v):
                if not w in visited:
                    push(frontier, (G[v][w].get(weight, 1), next(c), v, w))
            if data:
                yield u, v, G[u][v]
            else:
                yield u, v


def prim_mst(G, weight='weight'):
    """Return a minimum spanning tree or forest of an undirected
    weighted graph.

    A minimum spanning tree is a subgraph of the graph (a tree) with
    the minimum sum of edge weights.

    If the graph is not connected a spanning forest is constructed.  A
    spanning forest is a union of the spanning trees for each
    connected component of the graph.

    Parameters
    ----------
    G : NetworkX Graph

    weight : string
       Edge data key to use for weight (default 'weight').

    Returns
    -------
    G : NetworkX Graph
       A minimum spanning tree or forest.

    Examples
    --------
    >>> G=nx.cycle_graph(4)
    >>> G.add_edge(0,3,weight=2) # assign weight 2 to edge 0-3
    >>> T=nx.prim_mst(G)
    >>> print(sorted(T.edges(data=True)))
    [(0, 1, {}), (1, 2, {}), (2, 3, {})]

    Notes
    -----
    Uses Prim's algorithm.

    If the graph edges do not have a weight attribute a default weight of 1
    will be used.
    """

    T = nx.Graph(nx.prim_mst_edges(G, weight=weight, data=True))
    # Add isolated nodes
    if len(T) != len(G):
        T.add_nodes_from([n for n, d in G.degree().items() if d == 0])
    # Add node and graph attributes as shallow copy
    for n in T:
        T.node[n] = G.node[n].copy()
    T.graph = G.graph.copy()
    return T


from pycanvass.utilities import _version

__version__ = _version

__all__=["complexnetwork", "resiliency", "data_bridge", "data_visualization","blocks", "distributionsystem", "eventloop", "global_variables", "utilities"]


import json
import os
from pathlib import Path
import logging
from pycanvass.global_variables import *
import re
from pycanvass.blocks import *
from pycanvass.utilities import _banner
from pycanvass.distributionsystem import *
from pycanvass.resiliency import *
from pycanvass.global_variables import *
from pycanvass.data_bridge import *
from pycanvass.complexnetwork import *
from pycanvass.forecast import *
from pycanvass.data_visualization import *
from pycanvass.eventloop import *
import getpass


# Setup logging controls here
# ---------------------------
log_level = settings["event_log"]
log_level = log_level.upper()
logging.basicConfig(format='%(asctime)s : %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p',
                    filename='canvass_events.log', filemode='w', level=logging.DEBUG)

def _input_project_config_file():
    current_folder_path, current_folder_name = os.path.split(os.getcwd())
    default_json_file_name = current_folder_path + "\\" + current_folder_name + "\\project_config.json"
    default_json_file_name = Path(default_json_file_name)

    if default_json_file_name.exists():
        print("[i] Loaded default project configuration file: {}".format(default_json_file_name))
        project_config_path = default_json_file_name
        gv.filepaths["model"] = project_config_path
        return True
    else:
        print("[x] Default project configuration file not found.")
        project_config_path= input("\n[i] Please enter the path to the project file (*.JSON):\n")
        gv.filepaths["model"] = project_config_path
        return project_config_path
    
    gv.filepaths["model"] = project_config_path
    

def build_config_file():
    config_file_name = "auto_project_config.json"
    print("|----------------------------------------------|")
    print("| Building a simulation configuration file     |")
    print("|----------------------------------------------|")
    project_name = input("[1 of 9] Name of your project:     ")
    author_name = input("[2 of 9] Author:                   ")
    location = input("[3 of 9] Location:                 ")
    node_file = input("[4 of 9] Path to your node file:   ")
    edge_file = input("[5 of 9] Path to your edge file:   ")
    threat_file = input("[6 of 9] Path to your threat file: ")
    event_name = input("[7 of 9] Event to be simulated:    ")
    event_intensity = input("[8 of 9] Event intensity:          ")
    results_folder = input("[9 of 9] Results folder (Hit Enter for Default): ")

    try:
        if json.loads(data_string):
            print("[i] Your inputs were converted to a valid JSON file.")
            with open(config_file_name,'w+') as filename:
                m = json.dumps(data_string)
                filename.write(m)
    except:
        print("[1] Please use option 1 for now. Fixing it in the next update.")
        

def setup():
    _banner()
    try:
        if _input_project_config_file() is True:
            return 
        else:
            print("[x] Did not find a default project configuration file.\n\n")
            make_config_file = input("[?] Please make one of the following choices:\n"
                "[1] Provide path to project configuration folder\n"
                "[2] No, make a config file for me\n")

            if make_config_file.lstrip() == "1":
                config_file_path = input("[?] Please enter the full path to your configuration file:\n")
                
            elif make_config_file.lstrip() == "2":
                build_config_file()
            else:
                print("[x] Your input was not valid. Please retry.\n\n")
                setup()
    except KeyboardInterrupt:
        y_or_n = input("[x] It seems you wanted to terminate the project. Are you sure? [Y/N]")
        if y_or_n.lstrip() == "Y" or y_or_n.lstrip() == "y" or y_or_n.lstrip() == "Yes" or y_or_n.lstrip() == "yes":
            print("[i] Exiting pyCanvass. Bye.")
            sys.exit()




setup()








import csv
import logging
import pprint
import sys
import json
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pycanvass.complexnetwork as cn
import pycanvass.data_bridge as db
import pycanvass.data_visualization as dv
import pycanvass.global_variables as gv
import pycanvass.eventloop as el
from networkx import DiGraph

pp = pprint.PrettyPrinter(indent=4)

global edge_status_list
edge_status_list = {}

global edge_switches
edge_switches = {}

u_settings = open(gv.filepaths['user_preferences'])
settings = json.load(u_settings)
u_settings.close()


def interdependent_network_connectivity():
    interdependent_network_connectivity_dict = {}
    interdependent_network_connectivity_dict["transport"] = 0.063  # <-- todo: Hard-coded, needs implementation
    return interdependent_network_connectivity_dict


def upstream_edge_info(wg, n):
    upstream_edges = list(nx.edge_dfs(wg, n, orientation='reverse'))
    # print(upstream_edges)
    up_switch_dict = {}
    e_file = gv.filepaths["edges"]
    for u in upstream_edges:
        # edge info:
        p = u[0].lstrip()
        q = u[1].lstrip()
        edge_of_path_name_1 = p + "_to_" + event_intensity(n)
        edge_of_path_name_2 = q + "_to_" + p
        edge_search_result_1 = db._edge_search(edge_of_path_name_1)
        edge_search_result_2 = db._edge_search(edge_of_path_name_2)
        try:
            with open(e_file, 'r') as f:
                csvr = csv.reader(f)
                csvr = list(csvr)
                for row in csvr:
                    if row[0].lstrip() == edge_of_path_name_1 or row[0].lstrip() == edge_of_path_name_2:
                        if edge_search_result_1 != 0 or edge_search_result_2 != 0:
                            if row[1].lstrip() == "switch":
                                # print("[i] Upstream Switch: {}, Status: {}, Availability: {}".format(row[0], row[4],
                                # row[13]))
                                up_switch_dict[row[0]] = row[4]
                            if row[1].lstrip() == "transformer":
                                pass
                                # print(
                                # " [i] Upstream Transformer: {}, Status: {}, Availability: {}".format(row[0], row[4],
                                # row[13]))

        except:
            print("[x] Error in up-stream edge search.")
    return up_switch_dict


def lose_edge(from_node, to_node):
    # Normal graph
    g_normal = gv.graph_collection[0]
    cn.lat_long_layout(g_normal, show=True, save=False, title="Network in its current state")
    g_normal_damaged = nx.Graph()
    try:
        g_normal.remove_edge(from_node, to_node)
    except:
        print("[i] This edge in the normal graph may have been lost before. Moving on. ")
    g_normal_damaged.add_edges_from(g_normal.edges())
    cn.add_node_attr(g_normal_damaged)
    gv.graph_collection.append(g_normal_damaged)

    # Total graph
    g_total = gv.graph_collection[1]
    cn.lat_long_layout(g_total, show=True, save=False, title="All possible paths in the network")
    g_total_damaged = nx.DiGraph()
    try:
        g_total.remove_edge(from_node, to_node)
    except:
        print("[i] This edge in the complete graph may have been lost before. Moving on. ")
    g_total_damaged.add_edges_from(g_total.edges())
    cn.add_node_attr(g_total_damaged)
    gv.graph_collection.append(g_total_damaged)

    edge_of_path_name = from_node + "_to_" + to_node
    try:
        db.edit_edge_status(edge_of_path_name, set_status=0, availability=0)
    except:
        print("[i] Excel file could not be edited. It needs to be closed during the simulation.")

    return g_normal_damaged


def downstream_edge_info(wg, n):
    downstream_edges = list(nx.dfs_edges(wg, n))
    down_switch_dict = {}
    e_file = gv.filepaths["edges"]
    for d in downstream_edges:
        # edge info:
        p = d[0].lstrip()
        q = d[1].lstrip()
        edge_of_path_name_1 = p + "_to_" + q
        edge_of_path_name_2 = q + "_to_" + p
        edge_search_result_1 = db._edge_search(edge_of_path_name_1)
        edge_search_result_2 = db._edge_search(edge_of_path_name_2)
        # try:
        with open(e_file, 'r+') as f:
            csvr = csv.reader(f)
            csvr = list(csvr)
            for row in csvr:
                if row[0].lstrip() == edge_of_path_name_1 or row[0].lstrip() == edge_of_path_name_2:
                    if edge_search_result_1 != 0 or edge_search_result_2 != 0:
                        if row[1].lstrip() == "switch":
                            # print("[i] Downstream Switch: {}, Status: {}, Availability: {}".format(row[0], row[4],
                            # row[13]))
                            down_switch_dict[row[0]] = row[4]
                        if row[1].lstrip() == "transformer":
                            pass
                            # print("[i] Downstream Transformer: {}, Status: {}, Availability: {}".format(row[0],
                            # row[4],
                            # row[13]))

        # except:
        #     print("[x] Error in downstream edge search.")

    return down_switch_dict


def edge_info(edgename):
    """

    :param edgename:
    :return:
    """
    return


def reconfigure(node_to_restore, damaged_graph):
    """
    Algorithm: Kruskal Minimum Spanning Tree
    Damaged graph.
    Path search from node_to_restore, in complete graph to generators
    :return:
    """

    generator_nodes = []
    complete_graph = gv.graph_collection[3]
    undir_complete_graph = nx.Graph()  # complete_graph.to_undirected()
    e_file = gv.filepaths["edges"]
    # Calculate the impact on edge and attribute it to undir_complete_graph:
    complete_graph_edgelist = complete_graph.edges()
    temp_edge_impact_dict = {}
    with open("edge_risk_profile_during_reconfiguration.csv", "w+", newline='') as risk_profile_data:
        for m in complete_graph_edgelist:
            edge_of_path_name_1 = m[0] + "_to_" + m[1]
            edge_of_path_name_2 = m[1] + "_to_" + m[0]
            edge_search_result_1 = db._edge_search(edge_of_path_name_1)
            edge_search_result_2 = db._edge_search(edge_of_path_name_2)
            with open(e_file, 'r+') as f:
                csvr = csv.reader(f)
                csvr = list(csvr)

                for row in csvr:
                    if row[0].lstrip() == edge_of_path_name_1 or row[0].lstrip() == edge_of_path_name_2:
                        if edge_search_result_1 != 0:
                            impact = impact_on_edge(edge_of_path_name_1)
                            temp_edge_impact_dict[edge_of_path_name_1] = impact
                            undir_complete_graph.add_edge(m[0], m[1], weight=impact)
                            write_string = m[0] + "_to_" + m[1] + "," + str(impact) + "\n"
                            risk_profile_data.write(write_string)
                        elif edge_search_result_2 != 0:
                            impact = impact_on_edge(edge_of_path_name_2)
                            temp_edge_impact_dict[edge_of_path_name_1] = impact
                            undir_complete_graph.add_edge(m[1], m[0], weight=impact)
                            write_string = m[1] + "_to_" + m[0] + "," + str(impact) + "\n"
                            risk_profile_data.write(write_string)

    cn.add_node_attr(undir_complete_graph)

    for k, v in undir_complete_graph.nodes(data=True):
        if float(v['gen']) > 0 or v['gen'].lstrip() == "inf":
            generator_nodes.append(k)
    print("Available generator nodes are {}".format(generator_nodes))
    logging.info("Reconfiguration function called")
    logging.info("Available generator nodes are {}".format(generator_nodes))

    minimum_spanning_edges = nx.minimum_spanning_edges(undir_complete_graph, data=False)
    new_graph: DiGraph = nx.DiGraph()

    for m in minimum_spanning_edges:
        edge_of_path_name_1 = m[0] + "_to_" + m[1]
        edge_of_path_name_2 = m[1] + "_to_" + m[0]
        edge_search_result_1 = db._edge_search(edge_of_path_name_1)
        edge_search_result_2 = db._edge_search(edge_of_path_name_2)
        with open(e_file, 'r+') as f:
            csvr = csv.reader(f)
            csvr = list(csvr)

            for row in csvr:
                if row[0].lstrip() == edge_of_path_name_1 or row[0].lstrip() == edge_of_path_name_2:
                    try:
                        impact = temp_edge_impact_dict[edge_of_path_name_1]
                    except:
                        impact = temp_edge_impact_dict[edge_of_path_name_2]

                    if edge_search_result_1 != 0:
                        new_graph.add_edge(m[0], m[1], weight=float(impact))

                    elif edge_search_result_2 != 0:
                        new_graph.add_edge(m[1], m[0], weight=float(impact))

    print(new_graph.edges())
    cn.add_node_attr(new_graph)
    cn.lat_long_layout(new_graph, show=True, save=False, title="Minimum Spanning Tree Restoration")

    try:
        cyclic_path = list(nx.find_cycle(new_graph, node_to_restore, orientation='ignore'))
        print(cyclic_path)
    except nx.NetworkXNoCycle:
        print("[i] All loads could be picked without loop elimination.")


def restore(from_node, to_node, commit=False, control=False):
    """
    Change switch status is upstream and downstream edges.
    Run power flow to verify.

    Modeling Convention:
    If you are connecting it to a real-time device, such as RTDS, please remember that all switches are sorted
    alphabetically.

    """
    g2 = gv.graph_collection[1]
    edges = g2.edges()
    wg = nx.DiGraph()  # wg: working graph
    wg.add_edges_from(edges)
    try:
        down_switch_dict = downstream_edge_info(wg, from_node)
    except Exception as e:
        if KeyError:
            logging.info("From node was not found in the node file. May be a spelling error.")
            print("[x] 'From Node' was not correctly spelled when the 'reconfigure' function was called.")
            sys.exit()

    modified_down_switch_dict = {}
    up_switch_dict = upstream_edge_info(wg, from_node)
    print("Down switches from node:")
    pp.pprint(down_switch_dict)

    for k, v in down_switch_dict.items():
        mv = ''
        number_of_down_changes = 0
        number_of_up_changes = 0
        if v == '1' and number_of_down_changes == 0:
            modified_down_switch_dict[k] = mv

    print("Up switches from node:")
    pp.pprint(up_switch_dict)

    # edge_switches_reconfigured = {}
    # modified_v = ""
    # switching_operations = 0
    # for k, v in edge_switches:
    #     if v == "0":
    #         modified_v = "1"
    #         edge_switches_reconfigured[k] = mv
    #         switching_operations += 1
    #         print("[i] Switch {} needs to turned ON.")

    #         # search upstream switches, and turn them off.


def distant_between_two_points(p1, p2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    :return : Distance between two points in miles
    """
    lon1 = p1[0]
    lat1 = p1[1]

    lon2 = p2[0]
    lat2 = p2[1]

    # Convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    # Haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    c = 2 * np.arcsin(np.sqrt(a))
    # Radius of earth in kilometers is 6371
    km = 6371 * c
    return km/1.61


def primary_threat_anchor_of_node(node, ignore_multiple_threats_within="100"):
    """
    This method helps in finding out the threat nodes that will actually impact a certain node.
    :param node: instance of a node object
    :param ignore_multiple_threats_within: unit km, within this radius, multiple threat nodes are considered the same.
    :return: list of all threat anchors for the given node object, and assumed impact.
    """

    most_destructive_threat_anchor = ""
    strength = 1
    tmp_var = 100000  # <- any very large value
    proximal_threat_anchors = []  # <- todo: superposition impact of other threat nodes

    for k, value in gv.threat_dict.items():
        "Calculate distance between each threat node. "
        "Complexity: O(N^2)"  # <- todo: improve the runtime complexity

        distance = distant_between_two_points(tuple((float(value.lat), float(value.long))),
                                              tuple((float(node.lat), float(node.long))))

        if tmp_var > distance:
            most_destructive_threat_anchor = value.anchor
            strength = value.strength
            tmp_var = distance

        water_risk = value.water_risk
        fire_risk = value.fire_risk
        wind_risk = value.wind_risk
        seismic_risk = value.seismic_risk

    primary_threat_anchor = {"name": most_destructive_threat_anchor, "strength": strength, "distance": distance,
                             "water_risk": water_risk, "fire_risk": fire_risk, "wind_risk": wind_risk,
                             "seismic_risk": seismic_risk}
    logging.info("For node %s, the most significant threat anchor is %s" % (node.name, most_destructive_threat_anchor))

    return primary_threat_anchor


def primary_repair_base(node):
    """
    Primary repair base should be dependent on the type of event, and then sorted by distance from the node being
    studied.
    Suggested rules for choosing the primary repair base:
    - If the water risk is high at a particular node, the repair base with a lot of hand-pumps to extract out the water
    should be the primary repair base.
    - If wind risk is high, and node edges are connected by OH lines, a repair base with a lot of OH line, pole
    replacements, and crew_truck should be the primary reapir base.

    :param node:
    :return:
    """
    most_useful_repair_base = ""
    tmp_var = 100000  # <- any very large value
    repair_bases = {}

    repair_node_usefulness_score = 0

    for k, value in gv.repair_dict.items():
        "calculate distance between each threat node. "
        "Complexity: O(N^2)"  # <- todo: improve the runtime complexity

        distance = distant_between_two_points(tuple((float(value.lat), float(value.long))),
                                              tuple((float(node.lat), float(node.long))))

        threat_node = primary_threat_anchor_of_node(node)

        if float(threat_node["water_risk"]) > (float(node.water_cc) / 10.0) * float(settings["tolerances"][node.category.lstrip()]["water"]) and float(value.handpump) > 0:
            repair_node_usefulness_score += float(value.handpump)

        if float(threat_node["wind_risk"]) > (float(node.wind_cc) / 10.0) * float(settings["tolerances"][node.category.lstrip()]["wind"]) and float(value.line) > 0 and float(value.pole) > 0:
            repair_node_usefulness_score += (float(value.line) + float(value.pole) + float(value.switches) + float(value.mobile_genset))

        if float(threat_node["fire_risk"]) > (float(node.fire_cc) / 10.0) * float(settings["tolerances"][node.category.lstrip()]["fire"]) and float(value.transformer) > 0 and float(value.pole) > 0:
            repair_node_usefulness_score += (float(value.line) + float(value.transformer) + float(value.pole) + float(value.switches) + float(value.mobile_genset))

        if float(threat_node["seismic_risk"]) > (float(node.seismic_cc) / 10.0) * float(settings["tolerances"][node.category.lstrip()]["seismic"]) and float(value.mobile_genset) > 0 and float(value.pole) > 0 and float(value.line) > 0:
            repair_node_usefulness_score = (float(value.handpump) + float(value.line) + float(value.transformer) + float(value.pole) + float(value.switches) + float(value.mobile_genset))

        repair_bases[k] = repair_node_usefulness_score

    most_useful_repair_base = max(repair_bases.keys(), key=(lambda key: repair_bases[key]))
    max_usefulness_score = repair_bases[most_useful_repair_base]

    for kk, vv in repair_bases.items():
        repair_bases[kk] = vv/max_usefulness_score

    # Normalization to ensure all usefulness scores are reported at values less than or equal to 1
    most_useful_repair_base = max(repair_bases.keys(), key=(lambda key: repair_bases[key]))

    primary_repair_base = {"name": most_useful_repair_base, "distance": distance, "repair_bases": repair_bases}
    logging.info("For node %s, the most useful repair base is %s" % (node.name, primary_repair_base['name']))

    return primary_repair_base


def node_repairability(n):
    """
    This function helps calculate how easy is it for repair crew to navigate to damaged sites, and have accessibility
    to resources to quickly repair the damages.
    :param n:
    :return:
    """
    node = node_object_from_node_name(n)
    prb = primary_repair_base(node)
    inc = interdependent_network_connectivity()
    x = prb["distance"] * inc["transport"] * 1/np.exp(event_intensity(n))

    return x


def impact_on_nodes(obj_nodes):
    """

    :param obj_nodes:
    :return: Dictionary that has all node names and corresponding impact of an event on that node.
    """
    for n in obj_nodes:
        gv.node_risk_dict[n.name] = float(impact_on_node(n))

    return gv.node_risk_dict


def impact_on_node(node_name):
    """
    impact = (strength of event/distance from threat node)*risk of the node
    :return:
    """

    node_name = node_name.lstrip()
    node = node_object_from_node_name(node_name)

    threat_anchor = primary_threat_anchor_of_node(node)

    logging.info('Calculating impact of threat labeled %s on node %s' % (threat_anchor["name"], node.name))

    # DEFAULTS
    # ---------
    water_tolerance = settings["tolerances"][node.category.lstrip()]["water"]  # FT
    wind_tolerance = settings["tolerances"][node.category.lstrip()]["wind"]
    fire_tolerance = settings["tolerances"][node.category.lstrip()]["fire"]
    seismic_tolerance = settings["tolerances"][node.category.lstrip()]["seismic"]

    if node.category.lstrip() == "res":
        try:
            relative_importance = gv.node_res[node.name]
        except:
            relative_importance = 2
    elif node.category.lstrip() == "biz":
        try:
            relative_importance = gv.node_biz[node.name]
        except:
            relative_importance = 4
    elif node.category.lstrip() == "wlf":
        try:
            relative_importance = gv.node_wlf[node.name]
        except:
            relative_importance = 8
    elif node.category.lstrip() == "shl":
        try:
            relative_importance = gv.node_shl[node.name]
        except:
            relative_importance = 9
    elif node.category.lstrip() == "law":
        try:
            relative_importance = gv.node_law[node.name]
        except:
            relative_importance = 4
    elif node.category.lstrip() == "wat":
        try:
            relative_importance = gv.node_wat[node.name]
        except:
            relative_importance = 6
    elif node.category.lstrip() == "com":
        try:
            relative_importance = gv.node_com[node.name]
        except KeyError:
            relative_importance = 1
    elif node.category.lstrip() == "trn":
        try:
            relative_importance = gv.node_trn[node.name]
        except:
            relative_importance = 6
    elif node.category.lstrip() == "sup":
        try:
            relative_importance = gv.node_sup[node.name]
        except:
            relative_importance = 7
    else:
        try:
            relative_importance = gv.node_utl[node.name]
        except:
            relative_importance = 9

    # WATER

    water_risk_escalation = 1
    anticipated_water_risk = float(threat_anchor["water_risk"])
    if anticipated_water_risk > 0.95 * float(water_tolerance):
        water_risk_escalation = (anticipated_water_risk - 0.95 * float(water_tolerance)) * 0.1
    overall_water_risk = water_risk_escalation * (anticipated_water_risk / water_risk_escalation)
    if node.preexisting_damage.lstrip() == "1":
        overall_water_risk = overall_water_risk * 1.25

    # WIND

    wind_risk_escalation = 1
    anticipated_wind_risk = float(threat_anchor["wind_risk"])
    if anticipated_wind_risk > 0.90 * float(wind_tolerance):
        wind_risk_escalation = (anticipated_wind_risk - 0.90 * float(wind_tolerance)) * 0.1
    overall_wind_risk = wind_risk_escalation * (anticipated_water_risk / water_risk_escalation)
    if node.preexisting_damage.lstrip() == "1":
        overall_wind_risk = overall_wind_risk * 1.25

    # FIRE

    fire_risk_escalation = 1
    anticipated_fire_risk = float(threat_anchor["fire_risk"])
    if anticipated_fire_risk > 0.98 * float(fire_tolerance):
        fire_risk_escalation = (anticipated_fire_risk - 0.98 * float(fire_tolerance)) * 0.1
    overall_fire_risk = fire_risk_escalation * (anticipated_water_risk / water_risk_escalation)
    if node.preexisting_damage.lstrip() == "1":
        overall_fire_risk = overall_fire_risk * 2.0

    # SEISMIC

    seismic_risk_escalation = 1
    anticipated_seismic_risk = float(threat_anchor["seismic_risk"])
    if anticipated_seismic_risk > 0.90 * float(seismic_tolerance):
        seismic_risk_escalation = (anticipated_water_risk - 0.95 * float(seismic_tolerance)) * 0.1
    overall_seismic_risk = seismic_risk_escalation * (anticipated_water_risk / water_risk_escalation)
    if node.preexisting_damage.lstrip() == "1":
        overall_seismic_risk = overall_seismic_risk * 2.5

    total_risk = (relative_importance * float(node.foliage) * (
            overall_wind_risk / float(node.wind_cc) + overall_water_risk / float(
        node.water_cc) + overall_fire_risk / float(node.fire_cc) + overall_seismic_risk / float(node.seismic_cc))) / (
                     float(threat_anchor['distance']))

    return total_risk


def node_resiliency(n, verbose=False):
    """
    Ability to withstand an impact, and recover in multiple ways.
    :return:
    """
    node = node_object_from_node_name(n)
    graph = gv.graph_collection[1]
    graph = graph.to_undirected()
    generator_nodes = []
    path_counter = 0
    for k, v in graph.nodes(data=True):
        if float(v['gen']) > 0 or v['gen'].lstrip() == "inf":
            generator_nodes.append(k)
    # print("[i] Available generator nodes that can be used to restore {} are {}".format(n, generator_nodes))
    for g in generator_nodes:
        path = nx.all_simple_paths(graph, n, g)
        path_counter += len(list(path))

    node_res = path_counter

    if float(node.backup_dg.lstrip()) > 0.01:
        node_res += float(node.backup_dg.lstrip())

    threat_anchor = primary_threat_anchor_of_node(node)

    if float(threat_anchor["wind_risk"]) > 75:
        if node_res != 0:
            node_res = node_res * float(node.wind_cc.lstrip()) / 10.0
        else:
            node_res = float(node.wind_cc.lstrip()) / 10.0

    if float(threat_anchor["water_risk"]) > 2.0:
        if node_res != 0:
            node_res = node_res * float(node.water_cc.lstrip()) / 10.0
        else:
            node_res = float(node.water_cc.lstrip()) / 10.0

    if float(threat_anchor["seismic_risk"]) > 5:
        if node_res != 0:
            node_res = node_res * float(node.seismic_cc.lstrip()) / 10.0
        else:
            node_res = float(node.seismic_cc.lstrip()) / 10.0

    if float(threat_anchor["fire_risk"]) > 150:
        if node_res != 0:
            node_res = node_res * float(node.fire_cc.lstrip()) / 10.0
        else:
            node_res = float(node.fire_cc.lstrip()) / 10.0

    if verbose:
        print("-------------------------")
        print("Node name    : {}".format(n))
        print("Redundancy   : {}".format(path_counter))
        print("Backup       : {}".format(node.backup_dg.lstrip()))
        print("Fire Proof   : {}".format(node.fire_cc.lstrip()))
        print("Water Proof  : {}".format(node.water_cc.lstrip()))
        print("Wind Proof   : {}".format(node.wind_cc.lstrip()))
        print("Seimic Tol.  : {}".format(node.seismic_cc.lstrip()))
        print("--------------------------")
        print("RESILIENCY   : {}".format(node_res))
        print("--------------------------\n")

    return node_res


def path_search(G, n1, n2, criterion="least_risk"):
    """

    :param G: Graph object
    :param n1: source node in a graph object
    :param n2: sink node in a graph object
    :return: path:
    """
    if criterion == "least_risk":
        path = nx.dijkstra_path(G, n1, n2, 'impact_on_edge')
    load_and_demand_query(path)
    return path


def find_spanning_tree():
    """
    Cover maximum nodes in a damaged network, using least number of edges, or such that the lowest sum of edges are achieved.
    Implementation of Kruskal's Algorithm.
    Validated.
    :return:
    """

    pass


def load_and_demand_query(path, verbose=True):
    loads = 0
    gens = 0
    demand_kw = 0.0
    gen_kw = 0.0
    e_file = gv.filepaths["edges"]
    if type(path) is str:
        n = node_object_from_node_name(path)
    else:
        for p in path:
            n = node_object_from_node_name(p)
            i = path.index(p)
            if i < len(path) - 1:
                q = path[i + 1]
                edge_of_path_name_1 = p + "_to_" + q
                edge_of_path_name_2 = q + "_to_" + p
                edge_search_result_1 = db._edge_search(edge_of_path_name_1)
                edge_search_result_2 = db._edge_search(edge_of_path_name_2)
                try:
                    with open(e_file, 'r+') as f:
                        csvr = csv.reader(f)
                        csvr = list(csvr)
                        for row in csvr:
                            if row[0].lstrip() == edge_of_path_name_1 or row[0].lstrip() == edge_of_path_name_2:
                                if edge_search_result_1 != 0 or edge_search_result_2 != 0:
                                    edge_status_list[row[0]] = row[4]
                                    if row[1].lstrip() == "switch":
                                        # print("[i] Switch {} found!".format(row[0]))
                                        edge_switches[row[0]] = row[4]
                except:
                    print("[x] Edge could not be queried for status for the nodes.")

            if int(n.load) > 0:
                loads += 1
                demand_kw += float(n.load)
            if n.gen.lstrip() != "inf":
                gen_kw += float(n.gen)

    if verbose:
        print("[i] Number of loads = {}, Demand = {} kW, Generation = {} kW".format(loads, demand_kw, gen_kw))
        print("[i] Edge Statuses of the path:")
        pp.pprint(edge_status_list)

        if verbose:
            print("[i] Switch Status:")
            pp.pprint(edge_switches)

    load_and_demand_query_dict = {}
    load_and_demand_query_dict["edge_status"] = edge_status_list
    load_and_demand_query_dict["switch_status"] = edge_switches

    return load_and_demand_query_dict


def adjusted_node_risk(node_risk, adjustment_factor):
    """
    :param node_risk:
    :param new_information:
    :return:
    """
    return node_risk * adjustment_factor


def node_threat(n_risk, cta, threat_type="Generic", threat_horizon="day", threat_credibility="1"):
    """
    This function calculates the threat at each node
    :param threat_horizon: options: day (default, hour, two_day, week, two_week, month, few-months, year)
    :param threat_credibility: min: 0, max: 1 (default).
    :param n_risk:
    :param threat_type: default value is Generic, other possible values can be based on event.
    :param cta: closest_threat_anchor: A tuple containing lat, long of a threat anchor
    :return: node_threat_value
    """
    node_threat_value = 0
    return node_threat_value


def node_risk_estimator(n, overall_bias=1):
    """
    Risk Formula = max_risk - (wind_risk_factor*Wind_CC +
                               water_risk_factor*Water_CC +
                               fire_risk_factor*Fire_CC +
                               eq_risk_factor*Seismic_CC)/40

    where, CC = Code Conformity
    :param n: Row from the CSV file that has all the information about the node
    :param overall_bias: additional control variable to customize the description of impact of an event on a node.
           For example, for all hospitals might be at lower risk due to last minute efforts,
            we can say overall_bias < 1, e.g. 0.4, but typically n[15] and overall_bias values are kept at 1
    :return: risk of a node, based on its location and conformity to construction standards
             1 means maximum risk, 0 means minimum risk.
             Lower value is clearly better.
    """
    return 1 - overall_bias * float(n[15].lstrip()) * (gv.event["wind_risk"] * float(n[11].lstrip()) +
                                                       gv.event["water_risk"] * float(n[12].lstrip()) +
                                                       gv.event["fire_risk"] * float(n[14].lstrip()) +
                                                       gv.event["seismic_risk"] * float(n[13].lstrip())) / 40.0


def impact_on_edges():
    pass


def event_intensity(node_name):
    """

    :return: event_intensity
    """

    node = node_object_from_node_name(node_name)
    threat_node = primary_threat_anchor_of_node(node)

    wind_event_intensity = 4
    fire_event_intensity = 1
    seismic_event_intensity = 2
    water_event_intensity = 2

    if float(threat_node["wind_risk"]) > 75 and float(threat_node["wind_risk"]) < 100:
        wind_event_intensity = 6
    elif float(threat_node["wind_risk"]) >= 100 and float(threat_node["wind_risk"]) < 120:
        wind_event_intensity = 8
    elif float(threat_node["wind_risk"]) >= 120:
        wind_event_intensity = 10

    if float(threat_node["fire_risk"]) > 150 and float(threat_node["fire_risk"]) < 200 and float(threat_node["duration"]) >= 1:
        fire_event_intensity = 6
    elif float(threat_node["fire_risk"]) > 150 and float(threat_node["fire_risk"]) < 200 and float(threat_node["duration"]) >= 2:
        fire_event_intensity = 8
    elif float(threat_node["fire_risk"]) > 200 and float(threat_node["duration"]) >= 2:
        fire_event_intensity = 10

    if float(threat_node["seismic_risk"]) >= 5 and float(threat_node["seismic_risk"]) < 6:
        seismic_event_intensity = 6
    elif float(threat_node["seismic_risk"]) >= 6 and float(threat_node["seismic_risk"]) < 7.5:
        seismic_event_intensity = 8
    elif float(threat_node["seismic_risk"]) >= 7.5:
        seismic_event_intensity = 10

    if float(threat_node["water_risk"]) > 2 and float(threat_node["water_risk"]) < 5:
        water_event_intensity = 6
    elif float(threat_node["water_risk"]) >= 5 and float(threat_node["water_risk"]) < 8:
        water_event_intensity = 8
    elif float(threat_node["water_risk"]) >= 8:
        water_event_intensity = 10

    ei = 0.25 * (wind_event_intensity+fire_event_intensity+seismic_event_intensity+water_event_intensity)

    return ei


def edge_object_from_edge_name(e_name):
    edge_names = [e.name for e in gv.obj_edges]
    if e_name in edge_names:
        i = edge_names.index(e_name)
        return gv.obj_edges[i]
    else:
        print("[x] %s could not be found or created as an edge object. Check your edge database, i.e. edge-file.csv or "
              "something similar." % e_name)
        return False


def impact_on_edge(e):
    """
    Risk = (foliage_risk*elemental_risk*event_intensity*max(risk_of_from_node, risk_of_to_node))/hardening
    :param e: Key of edge dictionary for global edge risk dictionary
    :return:
    """
    edge_name = e  # e[1].lstrip() + "_to_" + e[0].lstrip()
    edge = edge_object_from_edge_name(edge_name)
    elemental_risk = (float(edge.wind_risk) + float(edge.fire_risk) + float(edge.water_risk)) / 30.0
    foliage_risk = 1.0  # <-- todo: need to implement the foliage risk factor
    # where's the closest threat node?
    e_file = gv.filepaths["edges"]
    edge_search_result = db._edge_search(edge_name)
    hardening = 1
    try:
        with open(e_file, 'r+') as f:
            csvr = csv.reader(f)
            csvr = list(csvr)
            for row in csvr:
                if row[0].lstrip() == edge_name and edge_search_result != 0:
                    from_node_of_e = row[2].lstrip()
                    to_node_of_e = row[3].lstrip()
                    hardening = float(row[12])
    except:
        print("[x] Cannot find the requested edge.\n[i] Calculating impact on substation PCC instead.")
        from_node_of_e = "S1"
        to_node_of_e = "F1_1"
        return

    ev_intensity_from = float(event_intensity(from_node_of_e))
    ev_intensity_to = float(event_intensity(to_node_of_e))
    ev_intensity = 0.5 * (ev_intensity_from + ev_intensity_to)
    # primary_threat_anchor = primary_threat_anchor_of_node(from_node_of_e)
    risk_of_from_node = impact_on_node(from_node_of_e)
    risk_of_to_node = impact_on_node(to_node_of_e)
    x = (foliage_risk * ev_intensity * max(risk_of_from_node, risk_of_to_node)) / hardening
    return x




def nodal_calculations(graph, visualize=True, title=""):

    nodes = graph.nodes()
    sort_node_by_type()

    project_config_file = open(gv.filepaths["model"])
    project_settings = json.load(project_config_file)
    project_config_file.close()
    logging.info("Performing node risk calculation for all nodes")
    file_name = project_settings["project_name"] + "-nodal_calculation.csv"

    with open(file_name, 'w+') as node_file:
        node_file.write('name,lat,long,risk,resiliency,repairability\n')
        for k, v in nodes.items():
            x = impact_on_node(k.lstrip())
            y = node_resiliency(k.lstrip())
            z = node_repairability(k.lstrip())
            write_string = k + "," + v['lat'].lstrip() + "," + v['long'].lstrip() + "," + str(x) + "," + str(y) + "," + str(z) +"\n"
            node_file.write(write_string)


def node_object_from_node_name(n_name):
    """

    :param n_name:
    :return:
    """
    node_names = [n.name for n in gv.obj_nodes]
    if n_name in node_names:
        i = node_names.index(n_name)
        return gv.obj_nodes[i]
    else:
        print("[x] %s could not be found or created as a node object. Check your database." % n_name)
        return False


def sort_node_by_type():
    """
    Types of node: res = residential,
                   biz = business,
                   wlf = welfare,
                   shl = shelter,
                   law = law enforcement,
                   wat = water utility,
                   com = communication,
                   trn = transportation,
                   sup = supply chain,
                   src = generation node
                   sub = substation node, swing bus for the feeder
    :param type:
    :return:
    """
    node_file = gv.project["data"]["nodes"]

    with open(node_file) as f:
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)
        nodes = csv.reader(f)
        if has_header:
            next(nodes)  # Skip header row

        for node in nodes:
            if node[9].lstrip() == "res":
                gv.node_res[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "biz":
                gv.node_biz[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "wlf":
                gv.node_wlf[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "shl":
                gv.node_shl[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "law":
                gv.node_law[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "wat":
                gv.node_wat[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "com":
                gv.node_com[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "trn":
                gv.node_trn[node[0]] = node_risk_estimator(node)
            elif node[9].lstrip() == "sup":
                gv.node_sup[node[0]] = node_risk_estimator(node)
            else:
                gv.node_utl[node[0]] = node_risk_estimator(node)


def resiliency_downstream(graph, edgelist, event):
    """
    This function computes the resiliency of all downstream sections,
    mainly looking out for critical loads
    :param graph:
    :param edgelist: All downstream sections of the feeder, from perspective of an edge being analyzed
    :param event: dictionary describing the event magnitude
    :return:
    """
    print(edgelist)

    for e in edgelist:
        for oe in gv.obj_edges:
            if e[0].lstrip() == oe.from_node.lstrip() and e[1].lstrip() == oe.to_node.lstrip():
                gv.edge_risk_dict[tuple((e[0], e[1]))] = (event["water_risk"] * float(oe.water_risk.lstrip())
                                                          + event["wind_risk"] * float(oe.wind_risk.lstrip())
                                                          + event["fire_risk"] * float(oe.fire_risk.lstrip())) / 30

    # loads encountered

    # critical loads encountered

    # generators encountered

    # switches encountered

    pass


def resiliency_upstream(graph, edgelist, event):
    """
    This function computes the resiliency of all upstream sections,
    mainly looking out for generators
    :param edgelist: All upstream sections of the feeder, from persective of a feeder being analyzed
    :return:
    """
    print("Heat your ", edgelist)
    pass


def weigh_the_sections(graph, attr_name="impact_on_edge"):
    edgelist = graph.edges()
    for e in edgelist:
        edge_name = e[0].lstrip() + "_to_" + e[1].lstrip()
        x = impact_on_edge(edge_name)
        graph[e[0]][e[1]][attr_name] = x
        # print("[i] Weighing {} by anticipated event impact {}".format(edge_name, x))


def resiliency(analysis='nodal'):
    """
    This function helps compute the resiliency metric of the network of a
    node, or a network
    :param kwargs:
    :return:
    """

    # 0. what's the event?
    event = {"water_risk": 0.9, "wind_risk": 1.2, "fire_risk": 0.8}

    # 1. get the graphs

    g1 = gv.graph_collection[0]
    g2 = gv.graph_collection[1]
    nodes = g1.nodes()
    edges = g1.edges()
    print(nodes)
    print(edges)

    # 2. for which node and edge are you doing the analysis?
    wg = nx.DiGraph()  # wg: working graph
    wg.add_edges_from(edges)

    for n in nodes:
        # Finding Downstream Edges
        print("Downstream Edges of %s -->" % n)
        downstream_edges = list(nx.dfs_edges(wg, n))
        resiliency_downstream(g2, downstream_edges, event)

        # Finding Upstream Edges
        print("Upstream Edges of %s -->" % n)
        upstream_edges = list(nx.edge_dfs(wg, n, orientation='reverse'))
        resiliency_upstream(g2, upstream_edges, event)

    # 3. track back from that edge to a source, or multiple sources
